# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
The PDF preview panel.

The widget module contains the real widget, the documents module a simple
abstraction and caching of PDF documents with their filenames,
and the printing module contains code to print a PDF document, either
via a PostScript rendering or by printing raster images to a QPrinter.

All the point & click stuff is handled in the pointandclick module.

"""


import functools
import os
import platform
import sys
import weakref

from PyQt6.QtCore import QSettings, QTimer, Qt, pyqtSignal
from PyQt6.QtGui import QAction, QActionGroup, QColor, QKeySequence, QPalette
from PyQt6.QtWidgets import (
    QApplication, QComboBox, QLabel, QSpinBox, QWidgetAction)

import app
import actioncollection
import actioncollectionmanager
import icons
import job
import pagedview
import qpageview.document
import qutil
import panel
import listmodel
import gadgets.drag

from . import documents

# viewModes from qpageview:
from qpageview import FixedScale, FitWidth, FitHeight, FitBoth
from qpageview import Horizontal, Vertical


def activate(func):
    """Decorator for MusicViewPanel methods/slots.

    The purpose is to first activate the widget and only perform an action
    when the event loop starts. This gives the PDF widget the chance to resize
    and position itself correctly.

    """
    @functools.wraps(func)
    def wrapper(self):
        instantiated = bool(super(panel.Panel, self).widget())
        self.activate()
        if instantiated:
            func(self)
        else:
            QTimer.singleShot(0, lambda: func(self))
    return wrapper


class MusicViewPanel(panel.Panel):
    def __init__(self, mainwindow):
        super().__init__(mainwindow)
        self.toggleViewAction().setShortcut(QKeySequence("Meta+Alt+M"))
        mainwindow.addDockWidget(Qt.DockWidgetArea.RightDockWidgetArea, self)

        ac = self.actionCollection = Actions(self)
        actioncollectionmanager.manager(mainwindow).addActionCollection(ac)
        ac.music_print.triggered.connect(self.printMusic)
        ac.music_save_settings.triggered.connect(self.writeSettings)
        ac.music_maximize.triggered.connect(self.maximize)
        ac.music_jump_to_cursor.triggered.connect(self.jumpToCursor)
        ac.music_sync_cursor.triggered.connect(self.toggleSyncCursor)
        ac.music_copy_image.triggered.connect(self.copyImage)
        ac.music_copy_text.triggered.connect(self.copyText)
        ac.music_document_select.documentsChanged.connect(self.updateActions)
        ac.music_copy_image.setEnabled(False)
        ac.music_copy_text.setEnabled(False)
        ac.music_reload.triggered.connect(self.reloadView)
        ac.music_clear.triggered.connect(self.clearView)

        # load the state of the actions from the preferences
        s = QSettings()
        s.beginGroup("musicview")
        ac.music_sync_cursor.setChecked(s.value("sync_cursor", False, bool))
        props = pagedview.PagedView.properties().setdefaults().load(s)
        ac._viewActions.updateFromProperties(props)
        ac._viewActions.viewRequested.connect(self.widget)   # force creation

    def translateUI(self):
        self.setWindowTitle(_("window title", "Music View"))
        self.toggleViewAction().setText(_("&Music View"))

    def createWidget(self):
        from . import widget
        w = widget.MusicView(self)
        s = QSettings()
        s.beginGroup("musicview")
        w.view.readProperties(s)
        w.view.rubberband().selectionChanged.connect(self.updateSelection)
        self.actionCollection._viewActions.setView(w.view)
        selector = self.actionCollection.music_document_select
        selector.currentDocumentChanged.connect(w.openDocument)
        selector.documentClosed.connect(w.clear)

        if selector.currentDocument():
            # open a document only after the widget has been created;
            # this prevents many superfluous resizes
            def open():
                if selector.currentDocument():
                    w.openDocument(selector.currentDocument())
            QTimer.singleShot(0, open)
        return w

    def writeSettings(self):
        """Save the current view properties as default."""
        if self.instantiated():
            s = QSettings()
            s.beginGroup("musicview")
            self.widget().view.writeProperties(s)

    def updateSelection(self, rect):
        self.actionCollection.music_copy_image.setEnabled(bool(rect))
        self.actionCollection.music_copy_text.setEnabled(bool(rect))

    def updateActions(self):
        ac = self.actionCollection
        ac.music_print.setEnabled(bool(ac.music_document_select.documents()))

    @activate
    def printMusic(self):
        if self.widget().view.pageCount():
            self.widget().view.print()

    @activate
    def jumpToCursor(self):
        self.widget().showCurrentLinks(True, 10000)

    @activate
    def reloadView(self):
        d = self.mainwindow().currentDocument()
        group = documents.group(d)
        if group.update() or group.update(False):
            ac = self.actionCollection
            ac.music_document_select.setCurrentDocument(d)

    def clearView(self):
        """Clear the music view and 'forget' the compiled documents."""
        if self.instantiated():
            d = self.mainwindow().currentDocument()
            documents.group(d).clear()
            self.actionCollection.music_document_select.setCurrentDocument(d)
            self.widget().clear()

    def toggleSyncCursor(self):
        QSettings().setValue("musicview/sync_cursor",
            self.actionCollection.music_sync_cursor.isChecked())

    def copyImage(self):
        page, rect = self.widget().view.rubberband().selectedPage()
        if not page:
            return
        filename = self.widget().view.document().filename()
        import copy2image
        copy2image.copy_image(self, page, rect, filename)

    def copyText(self):
        text = self.widget().view.rubberband().selectedText()
        if text:
            QApplication.clipboard().setText(text)


class Actions(actioncollection.ActionCollection):
    name = "musicview"
    def createActions(self, panel):
        self._viewActions = va = pagedview.ViewActions()
        self.music_document_select = DocumentChooserAction(panel)
        self.music_print = QAction(panel)
        self.music_zoom_in = va.zoom_in
        self.music_zoom_out = va.zoom_out
        self.music_zoom_original = va.zoom_original
        self.music_zoom_combo = va.zoomer
        self.music_rotate_left = va.rotate_left
        self.music_rotate_right = va.rotate_right
        self.music_fit_width = va.fit_width
        self.music_fit_height = va.fit_height
        self.music_fit_both = va.fit_both
        self.music_single_pages = va.layout_single
        self.music_two_pages_first_right = va.layout_double_right
        self.music_two_pages_first_left = va.layout_double_left
        self.music_raster = va.layout_raster
        self.music_horizontal = va.horizontal
        self.music_vertical = va.vertical
        self.music_continuous = va.continuous
        self.music_save_settings = QAction(panel)
        self.music_maximize = QAction(panel)
        self.music_jump_to_cursor = QAction(panel)
        self.music_sync_cursor = QAction(panel, checkable=True)
        self.music_copy_image = QAction(panel)
        self.music_copy_text = QAction(panel)
        self.music_pager = va.pager
        self.music_next_page = va.next_page
        self.music_prev_page = va.previous_page
        self.music_magnifier = va.magnifier
        self.music_reload = QAction(panel)
        self.music_clear = QAction(panel)

        self.music_print.setIcon(icons.get('document-print'))
        self.music_maximize.setIcon(icons.get('view-fullscreen'))
        self.music_jump_to_cursor.setIcon(icons.get('go-jump'))
        self.music_copy_image.setIcon(icons.get('edit-copy'))
        self.music_copy_text.setIcon(icons.get('edit-copy'))
        self.music_clear.setIcon(icons.get('edit-clear'))

        self.music_document_select.setShortcut(QKeySequence(Qt.Modifier.SHIFT | Qt.Modifier.CTRL | Qt.Key.Key_O))
        self.music_jump_to_cursor.setShortcut(QKeySequence(Qt.Modifier.CTRL | Qt.Key.Key_J))
        self.music_copy_image.setShortcut(QKeySequence(Qt.Modifier.SHIFT | Qt.Modifier.CTRL | Qt.Key.Key_C))
        self.music_reload.setShortcut(QKeySequence(Qt.Key.Key_F5))

    def translateUI(self):
        self.music_document_select.setText(_("Select Music View Document"))
        self.music_print.setText(_("&Print Music..."))
        self.music_zoom_combo.setText(_("Zoom Music"))
        self.music_save_settings.setText(_("Save current View settings as default"))
        self.music_maximize.setText(_("&Maximize"))
        self.music_jump_to_cursor.setText(_("&Jump to Cursor Position"))
        self.music_sync_cursor.setText(_("S&ynchronize with Cursor Position"))
        self.music_copy_image.setText(_("Copy to &Image..."))
        self.music_copy_text.setText(_("Copy Selected &Text"))
        self.music_reload.setText(_("&Reload"))
        self.music_clear.setText(_("Clear"))


class ComboBoxAction(QWidgetAction):
    """A widget action that opens a combobox widget popup when triggered."""
    def __init__(self, panel):
        super().__init__(panel)
        self.triggered.connect(self.showPopup)

    def showPopup(self):
        """Called when our action is triggered by a keyboard shortcut."""
        # find the widget in our floating panel, if available there
        for w in self.createdWidgets():
            if w.window() == self.parent():
                w.showPopup()
                return
        # find the one in the main window
        for w in self.createdWidgets():
            if w.window() == self.parent().mainwindow():
                w.showPopup()
                return


class DocumentChooserAction(ComboBoxAction):
    """A ComboBoxAction that keeps track of the current text document.

    It manages the list of generated PDF documents for every text document.
    If the mainwindow changes its current document and there are PDFs to display,
    it switches the current document.

    It also switches to a text document if a job finished for that document,
    and it generated new PDF documents.

    """

    documentClosed = pyqtSignal()
    documentsChanged = pyqtSignal()
    currentDocumentChanged = pyqtSignal(qpageview.document.Document)

    def __init__(self, panel):
        super().__init__(panel)
        self._model = None
        self._document = None
        self._documents = []
        self._currentIndex = -1
        self._indices = weakref.WeakKeyDictionary()
        panel.mainwindow().currentDocumentChanged.connect(self.slotDocumentChanged)
        documents.documentUpdated.connect(self.slotDocumentUpdated)

    def createWidget(self, parent):
        w = DocumentChooser(parent)
        w.activated[int].connect(self.setCurrentIndex)
        if self._model:
            w.setModel(self._model)
        return w

    def slotDocumentChanged(self, doc):
        """Called when the mainwindow changes its current document."""
        # only switch our document if there are PDF documents to display
        if self._document is None or documents.group(doc).documents():
            self.setCurrentDocument(doc)

    def slotDocumentUpdated(self, doc, j):
        """Called when a Job, finished on the document, has created new PDFs."""
        # if result files of this document were already displayed, the display
        # is updated. Else the current document is switched if the document was
        # the current document to be engraved (e.g. sticky or master) and the
        # the job was started on this mainwindow
        import engrave
        mainwindow = self.parent().mainwindow()
        if (doc == self._document or
            (job.attributes.get(j).mainwindow == mainwindow and
             doc == engrave.engraver(mainwindow).document())):
            self.setCurrentDocument(doc)

    def setCurrentDocument(self, document):
        """Displays the DocumentGroup of the given text Document in our chooser."""
        prev = self._document
        self._document = document
        if prev:
            prev.loaded.disconnect(self.updateDocument)
            prev.closed.disconnect(self.closeDocument)
            self._indices[prev] = self._currentIndex
        document.loaded.connect(self.updateDocument)
        document.closed.connect(self.closeDocument)
        self.updateDocument()

    def updateDocument(self):
        """(Re)read the output documents of the current document and show them."""
        docs = self._documents = documents.group(self._document).documents()
        self.setVisible(bool(docs))
        self.setEnabled(bool(docs))

        # make model for the docs
        m = self._model = listmodel.ListModel([d.filename() for d in docs],
            display = os.path.basename, icon = icons.file_type)
        m.setRoleFunction(Qt.ItemDataRole.UserRole, lambda f: f)
        for w in self.createdWidgets():
            w.setModel(m)

        index = self._indices.get(self._document, 0)
        if index < 0 or index >= len(docs):
            index = 0
        self.documentsChanged.emit()
        self.setCurrentIndex(index)

    def closeDocument(self):
        """Called when the current document is closed by the user."""
        self._document = None
        self._documents = []
        self._currentIndex = -1
        self.setVisible(False)
        self.setEnabled(False)
        self.documentClosed.emit()
        self.documentsChanged.emit()

    def documents(self):
        return self._documents

    def setCurrentIndex(self, index):
        if self._documents:
            self._currentIndex = index
            p = QApplication.palette()
            if not self._documents[index].updated:
                color = qutil.mixcolor(QColor(Qt.GlobalColor.red), p.color(QPalette.ColorRole.Base), 0.3)
                p.setColor(QPalette.ColorRole.Base, color)
            for w in self.createdWidgets():
                w.setCurrentIndex(index)
                w.setPalette(p)
            self.currentDocumentChanged.emit(self._documents[index])

    def currentIndex(self):
        return self._currentIndex

    def currentDocument(self):
        """Returns the currently selected Music document (Note: NOT the text document!)"""
        if self._documents:
            return self._documents[self._currentIndex]


class DocumentChooser(QComboBox):
    def __init__(self, parent):
        super().__init__(parent)
        self.setSizeAdjustPolicy(QComboBox.SizeAdjustPolicy.AdjustToContents)
        self.setEditable(True)
        self.lineEdit().setReadOnly(True)
        self.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        app.translateUI(self)
        gadgets.drag.ComboDrag(self).role = Qt.ItemDataRole.UserRole

    def translateUI(self):
        self.setToolTip(_("Choose the PDF document to display."))
        self.setWhatsThis(_(
            "Choose the PDF document to display or drag the file "
            "to another application or location."))



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Delivers information about a document.
"""


import itertools
import functools
import os
import re
import collections

from PyQt6.QtCore import (
    QCoreApplication, QObject, QSettings, QThread, QTimer, QUrl, pyqtSignal)

import document
import qsettings
import ly.lex
import lydocinfo
import lydocument
import app
import fileinfo
import cursortools
import tokeniter
import plugin
import variables
import lilypondinfo
import signals


__all__ = ['docinfo', 'info', 'mode']


def info(doc):
    """Returns a DocumentInfo instance for the given Document."""
    return DocumentInfo.instance(doc)


def docinfo(doc):
    """Return a LyDocInfo instance for the document."""
    return info(doc).lydocinfo()


def lilyinfo(doc):
    """Return a LilyPondInfo instance for the given document."""
    return info(doc).lilypondinfo()


def music(doc):
    """Return a music.Document instance for the document."""
    return info(doc).music()


def mode(doc, guess=True):
    """Returns the type of the given document. See DocumentInfo.mode()."""
    return info(doc).mode(guess)


def defaultfilename(doc):
    """Return a default filename that could be used for the document.

    The name is based on the score's title, composer etc.

    """
    i = info(doc)
    m = i.music()
    import ly.music.items as mus

    # which fields (in order) to harvest:
    s = QSettings()
    custom = s.value("custom_default_filename", False, bool)
    template = s.value("default_filename_template", "{composer}-{title}", str)
    if custom:
        # Retrieve all field names from the template string
        fields = [m.group(1) for m in re.finditer(r'\{(.*?)\}', template)]
    else:
        fields = ('composer', 'title')

    d = {}
    for h in m.find(mus.Header):
        for a in h.find(mus.Assignment):
            for f in fields:
                if f not in d and a.name() == f:
                    n = a.value()
                    if n:
                        t = n.plaintext()
                        if t:
                            d[f] = t
    # make filenames
    for k in d:
        d[k] = re.sub(r'\W+', '-', d[k]).strip('-')

    if custom:
        for k in fields:
            template = str.replace(template, '{' + k + '}', d.get(k, 'unknown'))
        filename = template
    else:
        filename = '-'.join(d[k] for k in fields if k in d)
    if not filename:
        filename = doc.documentName()
    ext = ly.lex.extensions[i.mode()]
    return filename + ext


class DocumentInfo(plugin.DocumentPlugin):
    """Computes and caches various information about a Document.

    To reduce lag, DocumentInfo computations are processed in the
    background while the document is being edited. As a consequence
    of this, DocumentInfo methods may return slightly older cached
    data while edits are being processed. This usually does not
    cause problems in practice, but if you do need absolutely
    current data, you can connect to your instance's contentsChanged
    signal to be notified when this processing has finished.

    """
    def __init__(self, doc):
        self._workerActive = False
        self._documentChanged = False
        # _lydocinfo and _music will be populated the first time their
        # respective getter functions are called
        self._reset()
        if isinstance(doc, document.EditorDocument):
            doc.contentsChanged.connect(self._invalidate)
            doc.changesStopped.connect(self._processChanges)
            doc.closed.connect(self._reset)

    contentsChanged = signals.Signal()

    def _reset(self):
        """Clear cached data when the document is opened or closed."""
        self._waitForWorker()
        self._lydocinfo = None
        self._music = None

    def lydocinfo(self):
        """Return the lydocinfo instance for our document."""
        self._waitForWorker()
        if self._lydocinfo is None:
            self._lydocinfo = _Worker.lydocinfo(self.document())
        return self._lydocinfo

    def music(self):
        """Return the music.Document instance for our document."""
        self._waitForWorker()
        if self._music is None:
            self._music = _Worker.music(self.document())
        self._music.include_path = self.includepath()
        return self._music

    def mode(self, guess=True):
        """Returns the type of document ('lilypond, 'html', etc.).

        The mode can be set using the "mode" document variable.
        If guess is True (default), the mode is auto-recognized based on the contents
        if not set explicitly using the "mode" variable. In this case, this function
        always returns an existing mode.

        If guess is False, auto-recognizing is not done and the function returns None
        if the mode wasn't set explicitly.

        """
        mode = variables.get(self.document(), "mode")
        if mode in ly.lex.modes:
            return mode
        if guess:
            return self.lydocinfo().mode()

    def includepath(self):
        """Return the configured include path.

        A path is a list of directories.

        If there is a session specific include path, it is used.
        Otherwise the path is taken from the LilyPond preferences.

        Currently the document does not matter.

        """
        # get the global include path
        include_path = qsettings.get_string_list(
            QSettings(), "lilypond_settings/include_path")

        # get the session specific include path
        import sessions
        session_settings = sessions.currentSessionGroup()
        if session_settings and session_settings.value("set-paths", False, bool):
            sess_path = qsettings.get_string_list(session_settings, "include-path")
            if session_settings.value("repl-paths", False, bool):
                include_path = sess_path
            else:
                include_path = sess_path + include_path

        return include_path

    def jobinfo(self, create=False):
        """Returns a two-tuple(filename, includepath).

        The filename is the file LilyPond shall be run on. This can be the
        original filename of the document (if it has a filename and is not
        modified), but also the filename of a temporarily saved copy of the
        document.

        The includepath is the same as self.includepath(), but with the
        directory of the original file prepended, only if a temporary
        'scratchdir'-area is used and the document does include other files
        (and therefore the original folder should be given in the include
        path to LilyPond).

        """
        includepath = self.includepath()
        filename = self.document().url().toLocalFile()

        # Determine the filename to run the engraving job on
        if not filename or self.document().isModified():
            # We need to use a scratchdir to save our contents to
            import scratchdir
            scratch = scratchdir.scratchdir(self.document())
            if create:
                scratch.saveDocument()
            if filename:
                includepath.insert(0, os.path.dirname(filename))
            if create or (scratch.path() and os.path.exists(scratch.path())):
                filename = scratch.path()
        return filename, includepath

    def includefiles(self):
        """Returns a set of filenames that are included by this document.

        The document's own filename is not added to the set.
        The configured include path is used to find files.
        Included files are checked recursively, relative to our file,
        relative to the including file, and if that still yields no file, relative
        to the directories in the includepath().

        This method uses caching for both the document contents and the other files.

        """
        return fileinfo.includefiles(self.lydocinfo(), self.includepath())

    def lilypondinfo(self):
        """Returns a LilyPondInfo instance that should be used by default to engrave the document."""
        version = self.lydocinfo().version()
        if version and QSettings().value("lilypond_settings/autoversion", False, bool):
            return lilypondinfo.suitable(version)
        return lilypondinfo.preferred()


    def child_urls(self):
        """Return a tuple of urls included by the Document.

        This only returns urls that are referenced directly, not searching
        via an include path. If the Document has no url set, an empty tuple
        is returned.

        """
        url = self.document().url()
        if url.isEmpty():
            return ()
        return tuple(url.resolved(QUrl(arg)) for arg in self.lydocinfo().include_args())

    def basenames(self):
        """Returns a list of basenames that our document is expected to create.

        The list is created based on include files and the define output-suffix and
        \bookOutputName and \bookOutputSuffix commands.
        You should add '.ext' and/or '-[0-9]+.ext' to find created files.

        """
        # if the file defines an 'output' variable, it is used instead
        output = variables.get(self.document(), 'output')
        filename = self.jobinfo()[0]
        if output:
            dirname = os.path.dirname(filename)
            return [os.path.join(dirname, name.strip())
                    for name in output.split(',')]

        mode = self.mode()

        if mode == "lilypond":
            return fileinfo.basenames(self.lydocinfo(), self.includefiles(), filename)

        elif mode == "html":
            pass

        elif mode == "texinfo":
            pass

        elif mode == "latex":
            pass

        elif mode == "docbook":
            pass

        return []

    def _invalidate(self):
        """Invalidate the cache when the document is changed.

        This forces the worker to regenerate cached data the next time
        _processChanges is called. The old data remains in the cache
        until the worker finishes, so DocumentInfo will return slightly
        outdated information while the worker is running. (This is the
        dirty secret to how we avoid lag.)

        """
        self._documentChanged = True

    def _processChanges(self):
        """Update cached data when the document is changed.

        This triggers a worker to regenerate _lydocinfo and _music
        in a background thread. These are slow operations, and
        processing them in the main thread caused significant lag
        when editing large LilyPond documents in older Frescobaldi
        versions (see issue #473).

        """
        # Note the worker operates on the live Document instance,
        # so it always sees the latest changes even if they happen
        # while the worker is running. There is no need to ever
        # retrigger an active worker; doing so will only cause more
        # of the very lag this design was intended to reduce.
        if self._documentChanged and not self._workerActive:
            try:
                worker = self._worker_instance
            except AttributeError:
                worker = self._worker_instance = _Worker(self.document())
                worker.moveToThread(_Worker.preferredThread())
                worker.finished.connect(self._slotWorkerFinished)

            self._workerActive = True
            # the worker modifies the document so we can't use a clone()
            self.document().moveToThread(worker.thread())
            QTimer.singleShot(0, worker.work)

    def _slotWorkerFinished(self, data):
        self._lydocinfo = data.lydocinfo
        self._music = data.music
        self._workerActive = False
        self._documentChanged = False
        self.contentsChanged.emit()

    def _waitForWorker(self):
        """Block (but keep the UI running) while the worker is active."""
        while self._workerActive:
            QCoreApplication.processEvents()


# named tuple type to pass data from the worker to the main thread
_WorkerData = collections.namedtuple("_WorkerData", "lydocinfo music")


class _Worker(QObject):
    """Worker to perform slow update operations in a separate thread."""
    def __init__(self, document):
        super().__init__()
        self._document = document
        self._documentThread = document.thread()

    def work(self):
        """Trigger this using a signal or QTimer to perform work."""
        lydocinfoData = self.lydocinfo(self._document)
        musicData = self.music(self._document)
        # return the document to its original thread
        self._document.moveToThread(self._documentThread)
        self.finished.emit(_WorkerData(lydocinfoData, musicData))

    @classmethod
    def preferredThread(cls):
        """Return the QThread where we want workers to live."""
        # all workers can share the same thread since we only have one
        # active document, and therefore one active worker, at a time
        try:
            thread = cls._thread
        except AttributeError:
            thread = cls._thread = QThread()
            thread.finished.connect(thread.deleteLater)
            thread.start()
        return thread

    # these are static methods so we can also call them directly from the
    # main thread as needed

    @staticmethod
    def lydocinfo(document):
        """Returns the data for DocumentInfo.lydocinfo()."""
        doc = lydocument.Document(document)
        v = variables.manager(document).variables()
        return lydocinfo.DocInfo(doc, v)

    @staticmethod
    def music(document):
        """Returns the data for DocumentInfo.music()."""
        import music
        doc = lydocument.Document(document)
        return music.Document(doc)

    finished = pyqtSignal("PyQt_PyObject")  # WorkerData



"""
MIDI input package
provides a dock which allows to capture midi events and insert notes

- input is static, not dynamic
- special midi events (e. g. damper pedal) can modify notes (e. g. duration)
  or insert elements (e. g. slurs)

current limitations:
- special events not implemented yet

TODO:
  dynamic input
"""

import re
from PyQt6.QtGui import QTextCursor

import time
import weakref

from PyQt6.QtCore import QObject, QThread, pyqtSignal

import app
import midihub
import midifile.event
import midifile.parser
import documentinfo

from . import elements

# What this does was originally undocumented. It appears intended to match
# chord and pitch names, but not commands or variables (thanks @ksnortum)
LY_REG_EXPR = re.compile(
    r'(?<![a-zA-Z#_^\-\\])[a-ps-zA-PS-Z]{1,3}(?![a-zA-Z])[\'\,]*'
    '|'
    r'(?<![<\\])<[^<>]*>(?!>)'
)

# Event codes from the MIDI specification
NOTE_OFF_EVENT = 8
NOTE_ON_EVENT = 9


class MidiIn:
    def __init__(self, widget):
        self._widget = weakref.ref(widget)
        self._portmidiinput = None
        self._listener = None
        self._chord = None

    def __del__(self):
        if isinstance(self._listener, Listener):
            self.capturestop()

    def widget(self):
        return self._widget()

    def open(self):
        # needed because close() does not reliably do its job
        midihub.refresh_ports()

        s = app.settings("midi")
        portname = s.value("input_port", midihub.default_input(), str)
        pollingtime = s.value("polling_time", 10, int)
        self._portmidiinput = midihub.input_by_name(portname)

        self._listener = Listener(self._portmidiinput, pollingtime)
        self._listener.receivedNoteEvent.connect(self.analyzeEvent)

    def close(self):
        if self._portmidiinput:
            self._portmidiinput.close()
        self._portmidiinput = None
        self._listener = None

    def capture(self):
        if not self._portmidiinput:
            self.open()
            if not self._portmidiinput:
                return
        doc = self.widget().mainwindow().currentDocument()
        self._language = documentinfo.docinfo(doc).language() or 'nederlands'
        self._activenotes = 0
        self._listener.start()

    def capturestop(self):
        self._listener.stop()
        if not self._listener.isFinished():
            self._listener.wait()
        self._activenotes = 0
        self.close()

    def analyzeEvent(self, event):
        if isinstance(event, midifile.event.NoteEvent):
            self.processNoteEvent(event.type, event.channel, event.note, event.value)

    def processNoteEvent(self, notetype, channel, notenumber, value):
        targetchannel = self.widget().channel()
        if targetchannel == 0 or channel == targetchannel - 1:  # '0' captures all
            # midi channels start at 1 for humans and 0 for programs
            if notetype == NOTE_ON_EVENT and value > 0: # value gives the velocity
                notemapping = elements.NoteMapping(self.widget().keysignature(), self.widget().accidentals() == 'sharps')
                note = elements.Note(notenumber, notemapping)
                if self.widget().chordmode():
                    if not self._chord:    # no Chord instance?
                        self._chord = elements.Chord()
                    self._chord.add(note)
                    self._activenotes += 1
                else:
                    self.addToDocument(note.output(self.widget().relativemode(), self._language))
            elif ((notetype == NOTE_OFF_EVENT
                   or (notetype == NOTE_ON_EVENT and value == 0))
                  and self.widget().chordmode()):
                self._activenotes -= 1
                if self._activenotes <= 0:    # activenotes could get negative under strange conditions
                    if self._chord:
                        self.addToDocument(self._chord.output(self.widget().relativemode(), self._language))
                    self._activenotes = 0    # reset in case it was negative
                    self._chord = None

    def addToDocument(self, text):
        view = self.widget().mainwindow()
        cursor = view.textCursor()

        if self.widget().repitchmode():
              music = cursor.document().toPlainText()[cursor.position() : ]
              notes = LY_REG_EXPR.search(music)
              if notes is not None:
                    start = cursor.position() + notes.start()
                    end = cursor.position() + notes.end()

                    cursor.beginEditBlock()
                    cursor.setPosition(start)
                    cursor.setPosition(end, QTextCursor.MoveMode.KeepAnchor)
                    cursor.insertText(text)
                    cursor.endEditBlock()

                    view.setTextCursor(cursor)

        else:
              # check if there is a space before cursor or beginning of line
              posinblock = cursor.position() - cursor.block().position()
              charbeforecursor = cursor.block().text()[posinblock-1:posinblock]

              if charbeforecursor.isspace() or cursor.atBlockStart():
                   cursor.insertText(text)
              else:
                   cursor.insertText(' ' +  text)


class Listener(QThread):
    receivedNoteEvent = pyqtSignal(midifile.event.NoteEvent)

    def __init__(self, portmidiinput, pollingtime):
        super().__init__()
        self._portmidiinput = portmidiinput
        self._pollingtime = pollingtime

    def run(self):
        self._capturing = True
        while self._capturing:
            while not self._portmidiinput.poll():
                time.sleep(self._pollingtime / 1000.)
                if not self._capturing:
                    break
            if not self._capturing:
                break
            data = self._portmidiinput.read(1)

            # midifile.parser.parse_midi_events is a generator which expects a long "byte string" from a file,
            # so we feed it one. But since it's just one event, we only need the first "generated" element.
            # First byte is time, which is unnecessary in our case, so we feed a dummy byte 77
            # and strip output by just using [1]. 77 is chosen randomly ;)
            s = bytearray([77, data[0][0][0], data[0][0][1], data[0][0][2], data[0][0][3]])
            event = next(midifile.parser.parse_midi_events(s))[1]

            if isinstance(event, midifile.event.NoteEvent):
                self.receivedNoteEvent.emit(event)

    def stop(self):
        self._capturing = False



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Keyboard part types.
"""


from PyQt6.QtWidgets import QGridLayout, QLabel, QSpinBox

import ly.dom
import ly.util

from . import _base
from . import register


class KeyboardPart(_base.PianoStaffPart):
    pass


class Piano(KeyboardPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Piano")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Piano", "Pno.")

    midiInstrument = 'acoustic grand'
    midiInstruments = (
        'acoustic grand',
        'bright acoustic',
        'electric grand',
        'honky-tonk',
    )


class ElectricPiano(KeyboardPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Electric piano")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Electric piano", "E.Pno.")

    midiInstrument = 'electric piano 1'
    midiInstruments = (
        'electric piano 1',
        'electric piano 2',
    )


class Harpsichord(KeyboardPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Harpsichord")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Harpsichord", "Hs.")

    midiInstrument = 'harpsichord'


class Clavichord(KeyboardPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Clavichord")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Clavichord", "Clv.")

    midiInstrument = 'clav'


class Organ(KeyboardPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Organ")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Organ", "Org.")

    midiInstrument = 'church organ'
    midiInstruments = (
        'drawbar organ',
        'percussive organ',
        'rock organ',
        'church organ',
        'reed organ',
    )

    def createWidgets(self, layout):
        super().createWidgets(layout)
        grid = layout.itemAt(layout.count() - 3).layout()
        self.pedalVoices = QSpinBox(minimum=0, maximum=4, value=1)
        self.pedalVoicesLabel = QLabel()
        self.pedalVoicesLabel.setBuddy(self.pedalVoices)
        grid.addWidget(self.pedalVoicesLabel, 2, 0)
        grid.addWidget(self.pedalVoices)

    def translateWidgets(self):
        super().translateWidgets()
        self.pedalVoicesLabel.setText(_("Pedal:"))
        self.pedalVoices.setToolTip(_(
            "Set to 0 to disable the pedal altogether."))

    def build(self, data, builder):
        super().build(data, builder)
        if self.pedalVoices.value():
            data.nodes.append(self.buildStaff(data, builder,
                'pedal', -1, self.pedalVoices.value(), clef="bass"))



class Celesta(KeyboardPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Celesta")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Celesta", "Cel.")

    midiInstrument = 'celesta'
    octave = 2
    transposition = (1, 0, 0)


class SynthPart(KeyboardPart):
    """Base class for synth parts.

    This is similar to _base.PianoStaffPart, except either
    upperVoices or lowerVoices can be set to zero, creating
    a single staff for writing monophonic lines.
    """
    # Derived classes should only include the subset of instruments
    # appropriate for that part type.
    # For example, SynthLead only includes the lead instruments.
    midiInstruments = (
        'synth bass 1',
        'synth bass 2',
        'synthstrings 1',
        'synthstrings 2',
        'synth voice',
        'synthbrass 1',
        'synthbrass 2',
        'lead 1 (square)',
        'lead 2 (sawtooth)',
        'lead 3 (calliope)',
        'lead 4 (chiff)',
        'lead 5 (charang)',
        'lead 6 (voice)',
        'lead 7 (fifths)',
        'lead 8 (bass+lead)',
        'pad 1 (new age)',
        'pad 2 (warm)',
        'pad 3 (polysynth)',
        'pad 4 (choir)',
        'pad 5 (bowed)',
        'pad 6 (metallic)',
        'pad 7 (halo)',
        'pad 8 (sweep)',
        'fx 1 (rain)',
        'fx 2 (soundtrack)',
        'fx 3 (crystal)',
        'fx 4 (atmosphere)',
        'fx 5 (brightness)',
        'fx 6 (goblins)',
        'fx 7 (echoes)',
        'fx 8 (sci-fi)',
    )

    # These can be set to zero to disable either staff
    minUpperVoices = 0
    minLowerVoices = 0

    def translateWidgets(self):
        super().translateWidgets()
        self.upperVoices.setToolTip(_(
            "Set to 0 to disable the right-hand part altogether."))
        self.lowerVoices.setToolTip(_(
            "Set to 0 to disable the left-hand part altogether."))


class SynthLead(SynthPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Synth lead")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Synth lead", "Syn.Ld.")

    # This is intended primarily for monophonic parts in treble clef,
    # so omit lower voices by default
    defaultLowerVoices = 0

    midiInstrument = 'lead 1 (square)'
    midiInstruments = (
        'lead 1 (square)',
        'lead 2 (sawtooth)',
        'lead 3 (calliope)',
        'lead 4 (chiff)',
        'lead 5 (charang)',
        'lead 6 (voice)',
        'lead 7 (fifths)',
        'lead 8 (bass+lead)',
    )


class SynthPad(SynthPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Synth pad")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Synth pad", "Syn.Pad")

    midiInstrument = 'pad 2 (warm)'
    midiInstruments = (
        'pad 1 (new age)',
        'pad 2 (warm)',
        'pad 3 (polysynth)',
        'pad 4 (choir)',
        'pad 5 (bowed)',
        'pad 6 (metallic)',
        'pad 7 (halo)',
        'pad 8 (sweep)',
    )


class SynthBass(SynthPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Synth bass")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Synth bass", "Syn.Bs.")

    # This is intended primarily for monophonic parts in bass clef,
    # so omit upper voices by default
    defaultUpperVoices = 0

    midiInstrument = 'synth bass 1'
    midiInstruments = (
        'synth bass 1',
        'synth bass 2',
    )


class SynthStrings(SynthPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Synth strings")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Synth strings", "Syn.Str.")

    midiInstrument = 'synthstrings 1'
    midiInstruments = (
        'synthstrings 1',
        'synthstrings 2',
    )


class SynthBrass(SynthPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Synth brass")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Synth brass", "Syn.Br.")

    midiInstrument = 'synthbrass 1'
    midiInstruments = (
        'synthbrass 1',
        'synthbrass 2',
    )

class SynthFx(SynthPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Synth effects")

    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Synth effects", "Syn.Fx.")

    midiInstrument = 'fx 3 (crystal)'
    midiInstruments = (
        'fx 1 (rain)',
        'fx 2 (soundtrack)',
        'fx 3 (crystal)',
        'fx 4 (atmosphere)',
        'fx 5 (brightness)',
        'fx 6 (goblins)',
        'fx 7 (echoes)',
        'fx 8 (sci-fi)',
    )


register(
    lambda: _("Keyboard instruments"),
    [
        Piano,
        ElectricPiano,
        Harpsichord,
        Clavichord,
        Organ,
        Celesta,
        SynthLead,
        SynthPad,
        SynthBass,
        SynthStrings,
        SynthBrass,
        SynthFx,
    ])





# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Base types for parts.
"""


import collections

from PyQt6.QtWidgets import (QCheckBox, QComboBox, QGridLayout, QHBoxLayout,
                             QLabel, QSpinBox)

import listmodel
import ly.dom


Category = collections.namedtuple("Category", "title items icon")


def translate(*args):
    """Translate the arguments using the application's language."""
    return _(*args)


class Base:
    """Base class for both Part and Container."""
    @staticmethod
    def title(_=translate):
        """Should return a title.

        If a translator is given, it is used instead of the builtin.

        """

    @staticmethod
    def short(_=translate):
        """Should return an abbreviated title.

        If a translator is given, it is used instead of the builtin.

        """

    def createWidgets(self, layout):
        """Should create widgets to adjust settings."""
        self.noSettingsLabel = QLabel()
        layout.addWidget(self.noSettingsLabel)

    def translateWidgets(self):
        """Should set the text in the widgets when the language changes."""
        self.noSettingsLabel.setText('({})'.format(_("No settings available.")))

    def accepts(self):
        """Should return a tuple of classes this part item accepts as child items."""
        return ()

    def build(self, data, builder):
        """Should populate the PartData (see build.py)."""
        data.nodes.append(ly.dom.Comment(f"Part {self.__class__.__name__}"))


class Part(Base):
    """Base class for Parts (that can't contain other parts)."""


class Container(Base):
    """Base class for "part" types that can contain others, like a Staff Group or Score, Book etc."""
    def accepts(self):
        return (Part, Container)


class Group(Container):
    """Base class for "part" types that are a group such as Book, BookPart and Score."""


# Mixin- or base classes with basic behaviour
class SingleVoicePart(Part):
    """Base class for a part creating a single single-voice staff."""
    midiInstrument = ''
    clef = None
    octave = 1
    transposition = None # or a three tuple (octave, note, alteration)

    def build(self, data, builder):
        a = data.assignMusic(None, self.octave)
        staff = ly.dom.Staff()
        builder.setInstrumentNamesFromPart(staff, self, data)
        if self.midiInstrument:
            builder.setMidiInstrument(staff, self.midiInstrument)
        seq = ly.dom.Seqr(staff)
        if self.clef:
            ly.dom.Clef(self.clef, seq)
        if self.transposition is not None:
            seq = builder.setStaffTransposition(seq, self.transposition)
        ly.dom.Identifier(a.name, seq)
        data.nodes.append(staff)


class PianoStaffPart(Part):
    """Base class for parts creating a piano staff."""
    midiInstruments = ()  # may contain a list of MIDI instruments.
    octave = 1  # for the right hand part; left is 1 octave lower.
    transposition = None

    # subclasses can override these as needed
    minUpperVoices = 1
    maxUpperVoices = 4
    defaultUpperVoices = 1
    minLowerVoices = 1
    maxLowerVoices = 4
    defaultLowerVoices = 1

    def createWidgets(self, layout):
        self.label = QLabel(wordWrap=True)
        self.upperVoicesLabel = QLabel()
        self.lowerVoicesLabel = QLabel()
        self.upperVoices = QSpinBox(minimum=self.minUpperVoices,
                                    maximum=self.maxUpperVoices,
                                    value=self.defaultUpperVoices)
        self.lowerVoices = QSpinBox(minimum=self.minLowerVoices,
                                    maximum=self.maxUpperVoices,
                                    value=self.defaultLowerVoices)
        self.dynamicsStaff = QCheckBox()
        self.dynamicsStaff.setChecked(True)

        self.upperVoicesLabel.setBuddy(self.upperVoices)
        self.lowerVoicesLabel.setBuddy(self.lowerVoices)

        layout.addWidget(self.label)
        grid = QGridLayout()
        grid.addWidget(self.upperVoicesLabel, 0, 0)
        grid.addWidget(self.upperVoices, 0, 1)
        grid.addWidget(self.lowerVoicesLabel, 1, 0)
        grid.addWidget(self.lowerVoices, 1, 1)
        layout.addLayout(grid)
        layout.addWidget(self.dynamicsStaff)

        if self.midiInstruments:
            self.createMidiInstrumentWidgets(layout)

        self.upperVoices.valueChanged.connect(self._voiceCountChanged)
        self.lowerVoices.valueChanged.connect(self._voiceCountChanged)
        self._voiceCountChanged()

    def createMidiInstrumentWidgets(self, layout):
        self.midiInstrumentLabel = QLabel()
        self.midiInstrumentSelection = QComboBox()
        self.midiInstrumentSelection.addItems(self.midiInstruments)
        self.midiInstrumentSelection.setCurrentIndex(
            self.midiInstruments.index(self.midiInstrument))
        box = QHBoxLayout()
        layout.addLayout(box)
        box.addWidget(self.midiInstrumentLabel)
        box.addWidget(self.midiInstrumentSelection)

    def translateWidgets(self):
        self.label.setText('{} <i>({})</i>'.format(
            _("Adjust how many separate voices you want on each staff."),
            _("This is primarily useful when you write polyphonic music "
              "like a fugue.")))
        self.upperVoicesLabel.setText(_("Right hand:"))
        self.lowerVoicesLabel.setText(_("Left hand:"))
        self.dynamicsStaff.setText(_("Center dynamics between staffs"))
        if self.midiInstruments:
            self.translateMidiInstrumentWidgets()

    def translateMidiInstrumentWidgets(self):
        self.midiInstrumentLabel.setText(_("MIDI instrument:"))

    def buildStaff(self, data, builder, name, octave, numVoices=1, node=None, clef=None):
        """Build a staff with the given number of voices and name."""
        staff = ly.dom.Staff(name, parent=node)
        if self.midiInstruments:
            midiInstrument = self.midiInstrumentSelection.currentText()
            if midiInstrument == 'percussive organ' and name != 'right':
                # The Hammond B3, which this MIDI instrument is intended
                # to emulate, only supports percussion on the upper manual
                midiInstrument = 'drawbar organ'
            builder.setMidiInstrument(staff, midiInstrument)
        else:
            builder.setMidiInstrument(staff, self.midiInstrument)
        c = ly.dom.Seqr(staff)
        if clef:
            ly.dom.Clef(clef, c)
        if self.transposition is not None:
            c = builder.setStaffTransposition(c, self.transposition)
        if numVoices == 1:
            a = data.assignMusic(name, octave)
            ly.dom.Identifier(a.name, c)
        else:
            c = ly.dom.Sim(c)
            for i in range(1, numVoices):
                a = data.assignMusic(name + ly.util.int2text(i), octave)
                ly.dom.Identifier(a.name, c)
                ly.dom.VoiceSeparator(c)
            a = data.assignMusic(name + ly.util.int2text(numVoices), octave)
            ly.dom.Identifier(a.name, c)
        return staff

    def buildDynamicsStaff(self, data, pianoSim):
        """Build a special staff to display dynamics."""
        dynamicsStaff = ly.dom.Dynamics(parent=pianoSim)
        a = data.assign('dynamics')
        ly.dom.Identifier(a.name, dynamicsStaff)
        s = ly.dom.Seq(a)
        ly.dom.Identifier(data.globalName, s).after = 1
        ly.dom.LineComment(_("Dynamics follow here."), s)
        ly.dom.BlankLine(s)

    def build(self, data, builder):
        """ Setup structure for a 1- or 2-staff PianoStaff. """
        upperCount = self.upperVoices.value()
        lowerCount = self.lowerVoices.value()
        if upperCount and lowerCount:
            # add two staves, with a respective number of voices.
            p = ly.dom.PianoStaff()
            s = ly.dom.Sim(p)
            self.buildStaff(data, builder, 'right', self.octave, upperCount, s)
            if self.dynamicsStaff.isChecked():
                # both staffs have to be present to use this feature
                self.buildDynamicsStaff(data, s)
            self.buildStaff(data, builder, 'left', self.octave - 1, lowerCount, s, "bass")
        elif upperCount:
            # add the treble staff only
            p = self.buildStaff(data, builder, None, self.octave, upperCount)
        elif lowerCount:
            # add the bass staff only
            p = self.buildStaff(data, builder, None, self.octave - 1, lowerCount, None, "bass")
        builder.setInstrumentNamesFromPart(p, self, data)
        data.nodes.append(p)

    def _voiceCountChanged(self, value=None):
        """Called when the number of voices in a staff is changed."""
        # Make sure we always have at least one voice present (this is mainly
        # for the synth parts that allow setting either count to zero)
        self.upperVoices.setMinimum(
            self.minUpperVoices if self.lowerVoices.value() else 1)
        self.lowerVoices.setMinimum(
            self.minLowerVoices if self.upperVoices.value() else 1)
        # Enable the option to center dynamics only when two staffs are present
        self.dynamicsStaff.setEnabled(
            self.upperVoices.value() and self.lowerVoices.value())


class ChordNames:
    def createWidgets(self, layout):
        self.chordStyleLabel = QLabel()
        self.chordStyle = QComboBox()
        self.chordStyleLabel.setBuddy(self.chordStyle)
        self.chordStyle.setModel(listmodel.ListModel(chordNameStyles, self.chordStyle,
            display=listmodel.translate))
        self.guitarFrets = QCheckBox()

        box = QHBoxLayout()
        box.addWidget(self.chordStyleLabel)
        box.addWidget(self.chordStyle)
        layout.addLayout(box)
        layout.addWidget(self.guitarFrets)

    def translateWidgets(self):
        self.chordStyleLabel.setText(_("Chord style:"))
        self.guitarFrets.setText(_("Guitar fret diagrams"))
        self.guitarFrets.setToolTip(_(
            "Show predefined guitar fret diagrams below the chord names "
            "(LilyPond 2.12 and above)."))
        self.chordStyle.model().update()

    def build(self, data, builder):
        p = ly.dom.ChordNames()
        a = data.assign('chordNames')
        ly.dom.Identifier(a.name, p)
        s = ly.dom.ChordMode(a)
        ly.dom.Identifier(data.globalName, s).after = 1
        i = self.chordStyle.currentIndex()
        if i > 0:
            ly.dom.Line('\\{}Chords'.format(
                ('german', 'semiGerman', 'italian', 'french')[i-1]), s)
        ly.dom.LineComment(_("Chords follow here."), s)
        ly.dom.BlankLine(s)
        data.nodes.append(p)
        if self.guitarFrets.isChecked():
            f = ly.dom.FretBoards()
            ly.dom.Identifier(a.name, f)
            data.nodes.append(f)
            data.includes.append("predefined-guitar-fretboards.ly")


chordNameStyles = (
    lambda: _("Default"),
    lambda: _("German"),
    lambda: _("Semi-German"),
    lambda: _("Italian"),
    lambda: _("French"),
)




# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Session menu.
"""


from PyQt6.QtGui import QActionGroup
from PyQt6.QtWidgets import QMenu

import app
import qutil
import sessions

from . import manager


class SessionMenu(QMenu):
    def __init__(self, mainwindow):
        super().__init__(mainwindow)
        app.translateUI(self)
        mgr = manager.get(mainwindow)
        ac = mgr.actionCollection
        ag = self._actionGroup = QActionGroup(self)
        ag.setExclusive(True)
        ag.addAction(ac.session_none)
        ag.triggered.connect(self.slotSessionsAction)
        self.addAction(ac.session_new)
        self.addAction(ac.session_save)
        self.addSeparator()
        self.addAction(ac.session_manage)
        self.addSeparator()
        self.addAction(ac.session_none)
        self.addSeparator()
        self.aboutToShow.connect(self.populate)
        self.groupMenus = []

    def translateUI(self):
        self.setTitle(_('menu title', '&Session'))

    def populate(self):
        groups = {}
        tl_sessions = []

        def add_session(menu, name, group=None):
            name = name.replace('&', '&&')
            fullname = group + '/' + name if group else name
            a = menu.addAction(name)
            a.setCheckable(True)
            if fullname == sessions.currentSession():
                a.setChecked(True)
                if group:
                    menu.setTitle('* ' + menu.title())
            a.setObjectName(fullname)
            self._actionGroup.addAction(a)

        def add_sessions():
            for name in tl_sessions:
                add_session(self, name)

        def add_groups():
            for k in sorted(groups.keys()):
                m = self.addMenu(k)
                self.groupMenus.append(m)
                for name in groups[k]:
                    add_session(m, name, k)
                qutil.addAccelerators(m.actions())

        def reset_menu():
            for m in self.groupMenus:
                m.deleteLater()
            self.groupMenus = []
            ac = manager.get(self.parentWidget()).actionCollection
            ag = self._actionGroup
            for a in ag.actions():
                if a is not ac.session_none:
                    self.removeAction(a)
                    ag.removeAction(a)
            ac.session_none.setChecked(not sessions.currentSession())

        reset_menu()
        for name in sessions.sessionNames():
            if '/' in name:
                group, name = name.split('/')
                if group in groups:
                    groups[group].append(name)
                else:
                    groups[group] = [name]
            else:
                tl_sessions.append(name)
        add_groups()
        if groups:
            self.addSeparator()
        add_sessions()
        qutil.addAccelerators(self.actions())

    def slotSessionsAction(self, action):
        if action.objectName() in sessions.sessionNames():
            manager.get(self.parentWidget()).startSession(action.objectName())



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Translation of some parts of Qt (like dialog buttons) via our PO files.

The strings that are needed are in the i18n/messages.py file in the
project root directory.

"""



from PyQt6.QtCore import QCoreApplication, QTranslator

import app


_translator = None

class Translator(QTranslator):
    """Subclass of QTranslator that gets its messages via the _() function."""
    def translate(self, context, sourceText, disambiguation="", n=-1):
        #_debug(context, sourceText)
        return _(context, sourceText)


def installTranslator():
    """Install a QTranslator so Qt's own texts are also translated."""
    global _translator
    if _translator is not None:
        QCoreApplication.removeTranslator(_translator)
    _translator = Translator()
    QCoreApplication.installTranslator(_translator)

def initialize():
    # just install again on change, so the widgets get a LanguageChange event
    app.languageChanged.connect(installTranslator)
    installTranslator()

# DEBUG: show translatable Qt messages once
_debugmessages = set()
def _debug(context, sourceText):
    l = len(_debugmessages)
    _debugmessages.add((context, sourceText))
    if len(_debugmessages) > l:
        print('_' + repr((context, sourceText)))




# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Dialog to copy contents from PDF to a raster image.
"""


import collections
import os
import tempfile

from PyQt6.QtCore import QEvent, QSettings, QSize, Qt
from PyQt6.QtGui import QBitmap, QColor, QDoubleValidator, QImage, QRegion
from PyQt6.QtWidgets import (
    QApplication, QCheckBox, QComboBox, QDialog, QDialogButtonBox, QFileDialog,
    QGridLayout, QHBoxLayout, QLabel, QMessageBox, QPushButton, QVBoxLayout)

import app
import util
import qutil
import icons
import qpageview.backgroundjob
import qpageview.imageview
import qpageview.export
import widgets.colorbutton


def copy_image(parent_widget, page, rect=None, filename=None):
    """Shows the dialog to copy a PDF page to a raster image.

    If rect is given, only that part of the page is copied.

    """
    dlg = Dialog(parent_widget)
    dlg.show()
    dlg.setPage(page, rect, filename)
    dlg.finished.connect(dlg.deleteLater)


class Dialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._filename = None
        self._page = None
        self._rect = None
        self._exporter = None
        self.runJob = qpageview.backgroundjob.SingleRun()
        self.imageViewer = qpageview.imageview.ImageView()
        self.dpiLabel = QLabel()
        self.dpiCombo = QComboBox(insertPolicy=QComboBox.InsertPolicy.NoInsert, editable=True)
        self.dpiCombo.lineEdit().setCompleter(None)
        self.dpiCombo.setValidator(QDoubleValidator(10.0, 1200.0, 4, self.dpiCombo))
        # 96 dpi is a standard PC screen; 192 dpi is high-DPI or "Retina"
        # 300 dpi and up are print resolutions
        self.dpiCombo.addItems([format(i) for i in (96, 192, 300, 600, 1200)])

        self.colorCheck = QCheckBox(checked=False)
        self.colorButton = widgets.colorbutton.ColorButton()
        self.colorButton.setColor(QColor(Qt.GlobalColor.white))
        self.grayscale = QCheckBox(checked=False)
        self.crop = QCheckBox()
        self.antialias = QCheckBox(checked=True)
        self.scaleup = QCheckBox(checked=False)
        self.dragfile = QPushButton(icons.get("image-x-generic"), None, None)
        self.copyfile = QPushButton(icons.get('edit-copy'), None, None)
        self.dragdata = QPushButton(icons.get("image-x-generic"), None, None)
        self.copydata = QPushButton(icons.get('edit-copy'), None, None)
        self.buttons = QDialogButtonBox(QDialogButtonBox.StandardButton.Close)
        self.saveButton = self.buttons.addButton('', QDialogButtonBox.ButtonRole.ApplyRole)
        self.saveButton.setIcon(icons.get('document-save'))

        layout = QVBoxLayout()
        self.setLayout(layout)

        hlayout = QHBoxLayout()
        hlayout.addWidget(self.imageViewer)

        controls = QGridLayout()
        hlayout.addLayout(controls)
        controls.addWidget(self.dpiLabel, 0, 0)
        controls.addWidget(self.dpiCombo, 0, 1)
        colorLayout = QHBoxLayout()
        colorLayout.setContentsMargins(0, 0, 0, 0)
        colorLayout.addWidget(self.colorCheck)
        colorLayout.addWidget(self.colorButton)
        controls.addLayout(colorLayout, 2, 0, 1, 2)
        controls.addWidget(self.grayscale, 3, 0, 1, 2)
        controls.addWidget(self.crop, 4, 0, 1, 2)
        controls.addWidget(self.antialias, 5, 0, 1, 2)
        controls.addWidget(self.scaleup, 6, 0, 1, 2)
        controls.addWidget(self.copydata, 8, 0, 1, 2)
        controls.addWidget(self.copyfile, 9, 0, 1, 2)
        controls.addWidget(self.dragdata, 10, 0, 1, 2)
        controls.addWidget(self.dragfile, 11, 0, 1, 2)
        controls.setRowStretch(7, 1)

        layout.addLayout(hlayout)
        layout.addWidget(widgets.Separator())
        layout.addWidget(self.buttons)

        app.translateUI(self)
        self.readSettings()
        self.finished.connect(self.writeSettings)
        self.dpiCombo.editTextChanged.connect(self.updateExport)
        self.colorCheck.toggled.connect(self.updateExport)
        self.colorButton.colorChanged.connect(self.updateExport)
        self.grayscale.toggled.connect(self.updateExport)
        self.scaleup.toggled.connect(self.updateExport)
        self.crop.toggled.connect(self.updateExport)
        self.antialias.toggled.connect(self.updateExport)
        self.buttons.rejected.connect(self.reject)
        self.copydata.clicked.connect(self.copyDataToClipboard)
        self.copyfile.clicked.connect(self.copyFileToClipboard)
        self.dragdata.pressed.connect(self.dragData)
        self.dragfile.pressed.connect(self.dragFile)
        self.dragdata.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.dragfile.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.saveButton.clicked.connect(self.saveAs)
        qutil.saveDialogSize(self, "copy_image/dialog/size", QSize(480, 320))

    def translateUI(self):
        self.setCaption()
        self.dpiLabel.setText(_("DPI:"))
        self.colorCheck.setText(_("Background:"))
        self.colorButton.setToolTip(_("Paper Color"))
        self.grayscale.setText(_("Gray"))
        self.grayscale.setToolTip(_("Convert image to grayscale."))
        self.crop.setText(_("Auto-crop"))
        self.antialias.setText(_("Antialias"))
        self.scaleup.setText(_("Scale 2x"))
        self.scaleup.setToolTip(_(
            "Render twice as large and scale back down\n"
            "(recommended for small DPI values)."))
        self.dragdata.setText(_("Drag"))
        self.dragfile.setText(_("Drag File"))
        self.copydata.setText(_("&Copy"))
        self.copyfile.setText(_("Copy &File"))
        self.saveButton.setText(_("&Save As..."))
        self.imageViewer.setWhatsThis(_(
            #xgettext:no-python-format
            "<p>\n"
            "Clicking toggles the display between 100% size and window size. "
            "Drag to copy the image to another application. "
            "Drag with Ctrl (or {command}) to scroll a large image.\n"
            "</p>\n"
            "<p>\n"
            "You can also drag the small picture icon in the bottom right, "
            "which drags the actual file on disk, e.g. to an e-mail message.\n"
            "</p>").format(command="\u2318"))
        self.updateFileTypeUITexts()

    def updateFileTypeUITexts(self):
        """Update the texts in buttons that carry file type information.

        Called from translateUI() and from updateExport().

        """
        self.dragdata.setToolTip(_("Drag the image to embed in a document."))
        self.dragfile.setToolTip(_("Drag the image as a PNG file."))
        self.copydata.setToolTip(_("Copy the image to the clipboard."))
        self.copyfile.setToolTip(_("Copy the image to the clipboard as a PNG file."))

    def readSettings(self):
        s = QSettings()
        s.beginGroup('copy_image')
        # 300 dpi is standard print resolution
        self.dpiCombo.setEditText(s.value("dpi", "300", str))
        color = s.value("papercolor", QColor(), QColor)
        self.colorButton.setColor(color if color.isValid() else Qt.GlobalColor.white)
        self.colorCheck.setChecked(color.isValid())
        self.grayscale.setChecked(s.value("grayscale", False, bool))
        self.crop.setChecked(s.value("autocrop", False, bool))
        self.antialias.setChecked(s.value("antialias", True, bool))
        self.scaleup.setChecked(s.value("scaleup", False, bool))

    def writeSettings(self):
        s = QSettings()
        s.beginGroup('copy_image')
        s.setValue("dpi", self.dpiCombo.currentText())
        color = self.colorButton.color() if self.colorCheck.isChecked() else QColor()
        s.setValue("papercolor", color)
        s.setValue("grayscale", self.grayscale.isChecked())
        s.setValue("autocrop", self.crop.isChecked())
        s.setValue("antialias", self.antialias.isChecked())
        s.setValue("scaleup", self.scaleup.isChecked())

    def setCaption(self):
        if self._filename:
            filename = os.path.basename(self._filename)
        else:
            filename = _("<unknown>")
        title = _("Image from {filename}").format(filename = filename)
        self.setWindowTitle(app.caption(title))

    def setPage(self, page, rect, filename):
        page = page.copy()
        if page.renderer:
            page.renderer = page.renderer.copy()
        self._page = page
        self._rect = rect
        self._filename = filename
        self.setCaption()
        self.updateExport()

    def updateExport(self):
        e = self._exporter = qpageview.export.ImageExporter(self._page, self._rect)
        e.filename = self._filename

        # update the enabled state of buttons
        self.dpiCombo.setEnabled(e.supportsResolution)
        self.colorCheck.setEnabled(e.supportsPaperColor)
        self.colorButton.setEnabled(e.supportsPaperColor and self.colorCheck.isChecked())
        self.grayscale.setEnabled(e.supportsGrayscale)
        self.crop.setEnabled(e.supportsAutocrop)
        self.antialias.setEnabled(e.supportsAntialiasing)
        self.scaleup.setEnabled(e.supportsOversample)

        # update the preferences of the exporter
        if e.supportsResolution:
            e.resolution = float(self.dpiCombo.currentText() or '300')
        if e.supportsPaperColor and self.colorCheck.isChecked():
            e.paperColor = self.colorButton.color()
        if e.supportsGrayscale:
            e.grayscale = self.grayscale.isChecked()
        if e.supportsAntialiasing:
            e.antialiasing = self.antialias.isChecked()
        if e.supportsAutocrop:
            e.autocrop = self.crop.isChecked()
        if e.supportsOversample:
            e.oversample = 2 if self.scaleup.isChecked() else 1

        # disable button actions
        self.dragfile.setEnabled(False)
        self.copyfile.setEnabled(False)
        self.dragdata.setEnabled(False)
        self.copydata.setEnabled(False)
        self.saveButton.setEnabled(False)

        # set filetype info in button tooltips
        self.updateFileTypeUITexts()

        # run the export job in a background thread
        self.runJob(e.document, self.exportDone)
        self.setCursor(Qt.CursorShape.WaitCursor)

    def exportDone(self, document):
        self.unsetCursor()
        self.imageViewer.setDocument(document)
        if self.imageViewer.viewMode() is not qpageview.FitBoth:
            self.imageViewer.zoomNaturalSize()
        # re-enable button actions
        self.dragfile.setEnabled(True)
        self.copyfile.setEnabled(True)
        self.dragdata.setEnabled(True)
        self.copydata.setEnabled(True)
        self.saveButton.setEnabled(True)

    def copyDataToClipboard(self):
        self._exporter.copyData()

    def copyFileToClipboard(self):
        self._exporter.copyFile()

    def dragData(self):
        self._exporter.dragData(self)
        self.dragdata.setDown(False)

    def dragFile(self):
        self._exporter.dragFile(self)
        self.dragfile.setDown(False)

    def saveAs(self):
        typeFilter = _("PNG Image (*.png)")
        filename = self._exporter.suggestedFilename()
        filename = QFileDialog.getSaveFileName(self,
            _("Save Image As"), filename, filter=typeFilter)[0]
        if filename:
            try:
                self._exporter.save(filename)
            except OSError:
                QMessageBox.critical(self, _("Error"), _(
                    "Could not save the image."))





# This file is part of the Frescobaldi Extensions project,
# https://github.com/frescobaldi-extensions
#
# Copyright (c) 2018 by Urs Liska and others
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

# VBCL parser
# Andrew Bernard 2017

import re

# compile the match patterns
comment = re.compile(r"^\s*#")
nv_pair = re.compile(r"^(.*):\s+(.*)$")
long_text_start = re.compile(r"^(.*):\s+<")
long_text_end = re.compile(r"^\s*>")
list_items_start = re.compile(r"^(.*):\s+\[")
list_items_end = re.compile(r"^\s*\]")


def check_mandatory_keys(d, mandatory_keys):
    """Check if all mandatory keys are present. Raise a ValueError if not."""
    if mandatory_keys:
        missing = []
        for key in mandatory_keys:
            if key not in d.keys():
                missing.append(key)
        if missing:
            raise ValueError(
                _("VBCL Error: Missing mandatory key(s) '{keys}'").format(
                    keys=', '.join(missing)))


def set_defaults(d, defaults):
    """Ensure optional properties are set to default/'empty' values."""
    for key in defaults:
        d[key] = d.get(key, defaults[key])


def parse(lines, mandatory_keys, defaults):
    """Returns a dictionary corresponding to a parsed VBCL string list."""
    d = dict()
    it = iter(lines)

    try:
        while True:
            line = next(it)
            # comments - discard
            if comment.search(line):
                continue
            else:
                # long text
                m = long_text_start.search(line)
                if m:
                    text = ''
                    while True:
                        line = next(it)
                        if long_text_end.search(line):
                            d[m.group(1)] = text.strip('\n')
                            break
                        text += (line.strip(' '))
                        text += '\n'
                        continue
                else:
                    # list
                    m = list_items_start.search(line)
                    if m:
                        items = list()
                        while True:
                            line = next(it)
                            if list_items_end.search(line):
                                d[m.group(1)] = items
                                break
                            items.append(line.strip(' \n'))
                            continue
                    else:
                        # name value pair
                        m = nv_pair.search(line)
                        if m:
                            d[m.group(1).strip()] = m.group(2).strip()
    except StopIteration:
        pass

    check_mandatory_keys(d, mandatory_keys)
    set_defaults(d, defaults)
    return d


def parse_file(filename, mandatory_keys=None, defaults=None):
    """Returns a dictionary corresponding to a parsed VBCL config file.
    Raises an exception if the given file doesn't exist or isn't readable"""

    with open(filename, encoding='utf-8') as f:
        cfg_dict = parse(f.read().split('\n'), mandatory_keys, defaults)
        return cfg_dict



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

# Tab widget for the configuration of the font setting command

from PyQt6.QtCore import (
    QSettings,
)
from PyQt6.QtWidgets import (
    QAbstractButton,
    QButtonGroup,
    QCheckBox,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QRadioButton,
    QScrollArea,
    QTabWidget,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)

import app
import highlighter


class FontCommandWidget(QScrollArea):
    """Displays and manipulates a font setting command."""

    # Store templates as class variablese
    lilypond_template = """  #(define fonts
    (set-global-fonts
<<<fontdefinitions>>>
     #:factor (/ staff-height pt 20)
   ))"""
    lilypond_paper_block = """\\paper {
<<<command>>>
}"""
    oll_template = '\\useNotationFont {}"{}"'
    oll_properties = """\\with {
<<<properties>>>
} """

    def __init__(self, parent):
        super().__init__(parent)

        self._cmd = {
            'lily': '',
            'oll': ''
        }
        self._full_cmd = {
            'lily': '',
            'oll': ''
        }
        self.approach = 'lily'

        self.font_labels = {}

        mainWidget = QWidget()
        self.setWidgetResizable(True)
        self.setWidget(mainWidget)
        self.setFrameShape(QScrollArea.Shape.NoFrame)
        layout = QVBoxLayout()
        mainWidget.setLayout(layout)
        col_layout = QHBoxLayout()
        layout.addLayout(col_layout)
        layout.addStretch()

        # Left column holding the options
        self.option_widget = QWidget()
        opt_layout = QVBoxLayout()
        opt_layout.setContentsMargins(0, 0, 0, 0)
        self.option_widget.setLayout(opt_layout)
        col_layout.addWidget(self.option_widget)

        # TextEdit to display the generated command
        self.command_edit = ce = QTextEdit()
        ce.setReadOnly(True)
        col_layout.addWidget(self.command_edit)

        # Which text font families to integrate?
        self.family_group = QGroupBox()
        opt_layout.addWidget(self.family_group)
        family_layout = QGridLayout()
        self.family_group.setLayout(family_layout)
        self.cb_roman = QCheckBox()
        self.font_labels['roman'] = QLabel(parent.selected_font('roman'))
        family_layout.addWidget(self.cb_roman, 0, 0)
        family_layout.addWidget(self.font_labels['roman'], 0, 1)

        self.cb_sans = QCheckBox()
        self.font_labels['sans'] = QLabel(parent.selected_font('sans'))
        family_layout.addWidget(self.cb_sans, 1, 0)
        family_layout.addWidget(self.font_labels['sans'], 1, 1)

        self.cb_typewriter = QCheckBox()
        self.font_labels['typewriter'] = QLabel(
            parent.selected_font('typewriter')
        )
        family_layout.addWidget(self.cb_typewriter, 2, 0)
        family_layout.addWidget(self.font_labels['typewriter'], 2, 1)

        # Choice between traditional and openLilyLib approach
        self.approach_group = QGroupBox()
        approach_layout = QVBoxLayout()
        self.approach_group.setLayout(approach_layout)
        self.approach_tab = QTabWidget()
        approach_layout.addWidget(self.approach_tab)
        opt_layout.addWidget(self.approach_group)

        self.trad_widget = QWidget()
        trad_layout = QVBoxLayout()
        self.trad_widget.setLayout(trad_layout)
        self.approach_tab.addTab(self.trad_widget, "")

        self.oll_widget = QWidget()
        oll_layout = QVBoxLayout()
        self.oll_widget.setLayout(oll_layout)
        self.approach_tab.addTab(self.oll_widget, "")

        # Configure traditional approach
        self.cb_music = QCheckBox()
        self.font_labels['music'] = QLabel(parent.selected_font('music'))
        trad_music_layout = QHBoxLayout()
        trad_music_layout.addWidget(self.cb_music)
        trad_music_layout.addWidget(self.font_labels['music'])
        trad_layout.addLayout(trad_music_layout)

        self.cb_paper_block = QCheckBox()
        trad_layout.addWidget(self.cb_paper_block)
        trad_layout.addStretch()

        # Configure openLilyLib approach
        self.cb_oll_music = QCheckBox()
        self.cb_oll_music.setChecked(True)
        self.cb_oll_music.setEnabled(False)
        self.font_labels['oll_music'] = QLabel(parent.selected_font('music'))
        oll_music_layout = QHBoxLayout()
        oll_music_layout.addWidget(self.cb_oll_music)
        oll_music_layout.addWidget(self.font_labels['oll_music'])
        oll_layout.addLayout(oll_music_layout)
        self.cb_oll = QCheckBox()
        oll_layout.addWidget(self.cb_oll)
        self.cb_loadpackage = QCheckBox()
        oll_layout.addWidget(self.cb_loadpackage)
        self.cb_extensions = QCheckBox()
        oll_layout.addWidget(self.cb_extensions)
        # Configure handling of stylesheet
        self.stylesheet_group = QGroupBox()
        self.stylesheet_buttons = QButtonGroup()
        oll_layout.addWidget(self.stylesheet_group)
        stylesheet_layout = QVBoxLayout()
        self.stylesheet_group.setLayout(stylesheet_layout)
        self.style_buttons = [QRadioButton() for i in range(3)]
        for i in range(3):
            b = self.style_buttons[i]
            stylesheet_layout.addWidget(b)
            self.stylesheet_buttons.addButton(b)
            self.stylesheet_buttons.setId(b, i)
        self.le_stylesheet = QLineEdit()
        stylesheet_layout.addWidget(self.le_stylesheet)
        oll_layout.addStretch()
        # enable line edit when custom stylesheet is selected
        self.stylesheet_buttons.buttonClicked.connect(
            lambda: self.le_stylesheet.setEnabled(
                self.stylesheet_buttons.checkedId() == 2
            )
        )

        self.loadSettings()
        self.window().finished.connect(self.saveSettings)
        # Connect widgets that trigger re-generation of the command
        # Map widget base classes to signal names
        signal_map = {
            QAbstractButton: 'toggled',
            QButtonGroup: 'buttonClicked',
            QTabWidget: 'currentChanged',
            QLineEdit: 'editingFinished'
        }
        trigger_widgets = [
            self.cb_roman,
            self.cb_sans,
            self.cb_typewriter,
            self.cb_music,
            self.cb_paper_block,
        ]
        trigger_widgets.extend([
            self.approach_tab,
            self.cb_oll,
            self.cb_loadpackage,
            self.cb_extensions,
            self.stylesheet_buttons,
            self.le_stylesheet
        ])
        for w in trigger_widgets:
            # For the current widget determine a supported base class
            # and connect the appropriate signal.
            # NOTE: When a new widget is added to the list it has to be
            # ensured it has a supported base class.
            for base_class in signal_map:
                if isinstance(w, base_class):
                    signal = getattr(w, signal_map[base_class])
                    signal.connect(self.invalidate_command)
                    break

        app.translateUI(self)

    def translateUI(self):
        self.family_group.setTitle(_("Set font families"))
        self.cb_roman.setText(_("Roman"))
        self.cb_sans.setText(_("Sans"))
        self.cb_typewriter.setText(_("Typewriter"))

        self.approach_group.setTitle(_("Configure command generation"))
        self.approach_tab.setTabText(0, _("Traditional"))
        self.approach_tab.setTabToolTip(0, _(
            "Specify fonts using the setting in a \\paper block."
        ))
        self.approach_tab.setTabText(1, _("openLilyLib"))
        self.approach_tab.setTabToolTip(1, _(
            "Specify fonts using the setting using openLilyLib.\n"
            + "NOTE: This requires openLilyLib (oll-core)\n"
            + "and the 'notation-fonts' openLilyLib package."
        ))

        self.cb_music.setText(_("Set music font"))
        self.cb_paper_block.setText(_("Complete \\paper block"))
        self.cb_paper_block.setToolTip(_(
            "Wrap setting in a complete \\paper block.\n"
            + "If unchecked, generate the raw font setting command."
        ))

        self.cb_oll_music.setText(_("Set music font"))
        self.cb_oll_music.setToolTip(_(
            "Specify the music font.\n"
            + "This is a reminder only and can not be unchecked "
            + "because the openLilyLib approach necessarily sets "
            + "the music font."
        ))
        self.cb_oll.setText(_("Load openLilyLib"))
        self.cb_oll.setToolTip(_(
            "Load openLilyLib (oll-core) explicitly.\n"
            + "Uncheck if oll-core is already loaded elsewhere."
        ))
        self.cb_loadpackage.setText(_("Load notation-fonts package"))
        self.cb_loadpackage.setToolTip(_(
            "Load the notation-fonts package explicitly.\n"
            + "Uncheck if it is already loaded elsewhere."
        ))
        self.cb_extensions.setText(_("Load font extensions (if available)"))
        self.cb_extensions.setToolTip(_(
            "Ask for loading font extensions.\n"
            + "Note that *some* fonts provide additional features\n"
            + "(e.g. glyphs) that can be made available through an\n"
            + "extension stylesheet if provided."
        ))
        self.stylesheet_group.setTitle(_("Font stylesheet"))
        self.stylesheet_group.setToolTip(_(
            "Select alternative stylesheet.\n"
            + "Fonts natively supported by the notation-fonts\n"
            + "package provide a default stylesheet to adjust\n"
            + "LilyPond's visuals (e.g. line thicknesses) to the\n"
            + "characteristic of the music font.\n"
            + "Check 'No stylesheet' to avoid a preconfigured\n"
            + "stylesheet to customize the appearance manually,\n"
            + "or check 'Custom stylesheet' to load another stylesheet\n"
            + "in LilyPond's search path."
        ))
        self.style_buttons[0].setText(_("Default stylesheet"))
        self.style_buttons[1].setText(_("No stylesheet"))
        self.style_buttons[2].setText(_("Custom stylesheet"))

    def loadSettings(self):
        s = QSettings()
        s.beginGroup('document-fonts-dialog')
        self.cb_roman.setChecked(s.value('set-roman', False, bool))
        self.cb_sans.setChecked(s.value('set-sans', False, bool))
        self.cb_roman.setChecked(s.value('set-roman', False, bool))
        self.cb_music.setChecked(s.value('set-music', True, bool))
        self.approach_tab.setCurrentIndex(s.value('approach-index', 0, int))
        self.cb_paper_block.setChecked(s.value('set-paper-block', True, bool))
        self.cb_oll.setChecked(s.value('load-oll', True, bool))
        self.cb_loadpackage.setChecked(s.value('load-package', True, bool))
        self.cb_extensions.setChecked(s.value('font-extensions', False, bool))
        style_type = s.value('style-type', 0, int)
        self.style_buttons[style_type].setChecked(True)
        self.le_stylesheet.setText(s.value('font-stylesheet', '', str))
        self.le_stylesheet.setEnabled(self.style_buttons[2].isChecked())

    def saveSettings(self):
        s = QSettings()
        s.beginGroup('document-fonts-dialog')
        s.setValue('set-roman', self.cb_roman.isChecked())
        s.setValue('set-sans', self.cb_sans.isChecked())
        s.setValue('set-roman', self.cb_roman.isChecked())
        s.setValue('set-paper-block', self.cb_paper_block.isChecked())
        s.setValue('approach-index', self.approach_tab.currentIndex())
        s.setValue('set-music', self.cb_music.isChecked())
        s.setValue('load-oll', self.cb_oll.isChecked())
        s.setValue('load-package', self.cb_loadpackage.isChecked())
        s.setValue('font-extensions', self.cb_extensions.isChecked())
        s.setValue('style-type', self.stylesheet_buttons.checkedId())
        s.setValue('font-stylesheet', self.le_stylesheet.text())

    def command(self, approach='lily'):
        """Return the command as shown in the Font Command tab."""
        if not self._cmd[approach]:
            self.invalidate_command()
        return self._cmd[approach]

    def generate_lily_command(self):
        """
        Generate a font setting command in the traditional LilyPond way
        using #(define fonts ...) in a \\paper block.
        Returns a tuple with two versions of the command:
        - command as shown in the Font Command tab
          => with various filters applied
        - "full" command without the filters.
          => as used in the Font Preview, and maybe the Document wizard
        """
        # Definitions are initially handled as string lists,
        # and later joined to multiline strings.
        fontdefs = []
        full_fontdefs = []
        template = self.lilypond_template

        def add_font_def(k, name, checked):
            """
            Add a font entry to the commands when applicable.
            """
            font_entry = f'     #:{k} "{name}"'
            if checked:
                fontdefs.append(font_entry)
                full_fontdefs.append(font_entry)

        def font_defs():
            """Compose the font definitions list."""
            add_font_def(
                'music', self.window().selected_font('music'),
                self.cb_music.isChecked()
            )
            add_font_def(
                'brace', self.window().selected_font('brace'),
                self.cb_music.isChecked()
            )
            add_font_def(
                'roman', self.window().selected_font('roman'),
                self.cb_roman.isChecked()
            )
            add_font_def(
                'sans', self.window().selected_font('sans'),
                self.cb_sans.isChecked()
            )
            add_font_def(
                'typewriter',
                self.window().selected_font('typewriter'),
                self.cb_typewriter.isChecked()
            )
            return "\n".join(fontdefs), "\n".join(full_fontdefs)

        fontdefs, full_fontdefs = font_defs()
        cmd = template.replace('<<<fontdefinitions>>>', fontdefs)
        full_cmd = template.replace('<<<fontdefinitions>>>', full_fontdefs)
        if self.cb_paper_block.isChecked():
            cmd = self.lilypond_paper_block.replace('<<<command>>>', cmd)
        full_cmd = self.lilypond_paper_block.replace('<<<command>>>', full_cmd)

        return cmd, full_cmd

    def generate_oll_command(self):
        """
        Generate a font setting command using openLilyLib's notation-fonts
        package.
        Returns a tuple with two versions of the command:
        - command as shown in the Font Command tab
          => with various filters applied
        - "full" command without the filters.
          => as used in the Font Preview, and maybe the Document wizard
        """
        # Handled initially as string lists, later joined to multiline strings
        cmd = []
        full_cmd = []
        properties = []
        full_properties = []

        def add_property(k, v, checked, force=True):
            """
            Add a property entry to the full command
            and conditionally to the regular command.
            """
            property = f'  {k} = {v}'
            if checked:
                properties.append(property)
            if checked or force:
                full_properties.append(property)

        # Load openLilyLib
        oll_include = '\\include "oll-core/package.ily"'
        full_cmd.append(oll_include)
        if self.cb_oll.isChecked():
            cmd.append(oll_include)

        # Load the notation-fonts package
        package_include = '\\loadPackage notation-fonts'
        full_cmd.append(package_include)
        if self.cb_loadpackage.isChecked():
            cmd.append(package_include)

        # TODO: Support independent explicit brace font
        add_property(
            'brace', '"{}"'.format(self.window().selected_font('brace')), False
        )

        # Specify text fonts
        add_property(
            'roman',
            '"{}"'.format(self.window().selected_font('roman')),
            self.cb_roman.isChecked(),
            force=True
        )
        add_property(
            'sans',
            '"{}"'.format(self.window().selected_font('sans')),
            self.cb_sans.isChecked(),
            force=True
        )
        add_property(
            'typewriter',
            '"{}"'.format(self.window().selected_font('typewriter')),
            self.cb_typewriter.isChecked(),
            force=True
        )

        # Optionally load font extensions
        if self.cb_extensions.isChecked():
            add_property('extensions', '##t', True)

        # Handle font stylesheet
        style_type = self.stylesheet_buttons.checkedId()
        # style_type == 0 => default stylesheet, doesn't have to be written
        if style_type == 1:
            # Don't use a stylesheet
            add_property('style', 'none', True)
        elif style_type == 2:
            # Use custom stylesheet (must be findable of course)
            add_property(
                'style',
                f'"{self.le_stylesheet.text()}"',
                True
            )

        # Generate the \with clause ...
        full_properties = self.oll_properties.replace(
            '<<<properties>>>',
            '\n'.join(full_properties)
        )
        # ... conditionally for the regular command
        if properties:
            properties = self.oll_properties.replace(
                '<<<properties>>>',
                '\n'.join(properties)
            )
        else:
            properties = ''
        # Inject properties in the \useNotationFont command
        cmd.append(
            self.oll_template.format(
                properties, self.window().selected_font('music')
            ))
        full_cmd.append(
            self.oll_template.format(
                full_properties,
                self.window().selected_font('music')
            ))

        # Return regular and full command
        return "\n".join(cmd), "\n".join(full_cmd)

    def invalidate_command(self, _=None):
        """
        Regenerate both 'lily' and 'oll' versions of the regular and full
        font commands. Display the regular command in the textedit and
        trigger showing the sample in the music font tab.
        """
        self._cmd['lily'], self._full_cmd['lily'] = self.generate_lily_command()
        self._cmd['oll'], self._full_cmd['oll'] = (self.generate_oll_command())
        self.approach = (
            'lily' if self.approach_tab.currentIndex() == 0 else 'oll'
        )
        display_cmd = self._cmd[self.approach]
        self.command_edit.setPlainText(display_cmd)
        highlighter.highlight(self.command_edit.document())
        for k in self.font_labels:
            font_key = 'music' if k == 'oll_music' else k
            self.font_labels[k].setText(self.window().selected_font(font_key))
        self.window().show_sample()

    def full_cmd(self, approach='lily'):
        """Return the (cached) full command for the requested approach."""
        if not self._full_cmd[approach]:
            self.invalidate_command()
        return self._full_cmd[approach]



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Base ExtensionPanel for extensions
"""

import panel

from . import ExtensionMixin

class ExtensionPanel(panel.Panel):
    """Base class for an Extension Tool Panel.
    It is a lightweight layer around the regular panel.Panel,
    its only purpose being to act as a bridge to ExtensionWidget and
    descendants.
    """
    def __init__(self, extension, widget_class, dock_area):
        self._extension = extension
        if not issubclass(widget_class, ExtensionMixin):
            raise TypeError(_(
                "Extension panel widget class '{classname}' "
                "is not a subclass of ExtensionMixin.\n"
                "Please derive either from ExtensionWidget "
                "or add extensions.ExtensionMixin as a second base class."
            ).format(classname=widget_class.__name__))
        self._widget_class = widget_class
        super().__init__(extension.mainwindow())
        self.hide()
        self.mainwindow().addDockWidget(dock_area, self)

    def createWidget(self):
        """Create the panel's widget.
        *If* an ExtensionPanel is actually instantiated it also has
        information about it's widget class, which we use here.
        If the widget's class is not dre"""
        try:
            w = self._widget_class(self)
            if not hasattr(w, 'extension'):
                w.extension = lambda: self.extension()
            return w
        except Exception:
            # If the instantiation of the widget fails we create a nice
            # error message and return an "empty" FailedExtensionWidget instead.
            import sys
            import traceback
            from PyQt6.QtWidgets import QMessageBox
            import appinfo
            from extensions.widget import FailedExtensionWidget

            extension = self.extension()
            info = extension.infos()

            msg_box = QMessageBox()
            msg_box.setWindowTitle(appinfo.appname)
            msg_box.setIcon(QMessageBox.Icon.Critical)
            msg_box.setText(_("Extension Tool Panel failed to load"))
            info_text = _(
                "The Tool Panel provided by the extension\n\n  {name}\n\n"
                "failed to load due to an exception. "
                "Please contact the extension's maintainer(s)."
            ).format(name=extension.display_name() or extension.name())
            maintainers = info.get('maintainers', [])
            for m in maintainers:
                info_text += f"\n  - {m}"
            msg_box.setInformativeText(info_text)
            exc_info = sys.exc_info()
            msg_box.setDetailedText(
                ' '.join(traceback.format_exception(*exc_info)))
            msg_box.exec()

            return FailedExtensionWidget(self)

    def translateUI(self):
        self.setWindowTitle(self.extension().display_name())

    def extension(self):
        return self._extension

    def settings(self):
        return self.extension().settings()



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
UrlRequester, a lineedit with a Browse-button.
"""

import os

from PyQt6.QtCore import pyqtSignal
from PyQt6.QtWidgets import (
    QFileDialog,
    QHBoxLayout,
    QLineEdit,
    QToolButton,
    QWidget
)

import app
import icons


class UrlRequester(QWidget):
    """Shows a lineedit and a button to select a file or directory.

    The lineEdit, button, and fileDialog attributes represent their
    respective objects.

    """
    changed = pyqtSignal()
    editingFinished = pyqtSignal()

    def __init__(
        self,
        parent=None,
        fileMode=QFileDialog.FileMode.Directory,
        mustExist=False
    ):
        super().__init__(parent)

        self._fileDialog = None
        self._dialogTitle = None
        self._mustExist = mustExist
        self._originalPath = ''

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(2)
        self.setLayout(layout)

        self.lineEdit = QLineEdit()
        layout.addWidget(self.lineEdit)
        self.button = QToolButton(clicked=self.browse)
        layout.addWidget(self.button)

        self.lineEdit.textChanged.connect(self._changed)
        self.lineEdit.editingFinished.connect(self._editingFinished)
        self._browse_clicked = False
        self.setFileMode(fileMode)
        app.translateUI(self)

    def translateUI(self):
        self.button.setToolTip(_("Open file dialog"))

    def _changed(self):
        """
        Emit the `changed` signal.
        If the `mustExist` property is set validate the current value
        and color the line edit.
        """
        if self.mustExist():
            if os.path.exists(self.path()):
                self.lineEdit.setStyleSheet('')
            else:
                # TODO: Apply the "Error" color of the current theme
                self.lineEdit.setStyleSheet('color:red')
        self.changed.emit()

    def _editingFinished(self):
        """Emit the editingFinished signal - if validation passes.

        If the focus changes from the lineEdit to the fileDialog
        no signal is emitted.

        If mustExist=True and the file doesn't exist, reset to the value
        before editing and suppress the signal.

        Only emit the signal if the path has actually changed.
        """
        if self._browse_clicked:
            self.fileDialog().setDirectory(self.lineEdit.text())
        elif self.mustExist() and not os.path.exists(self.path()):
            self.setPath(self._originalPath)
        elif self.path() != self._originalPath:
            self._originalPath = self.path()
            self.editingFinished.emit()

    def fileDialog(self, create=False):
        """Returns the QFileDialog, if already instantiated.

        If create is True, the dialog is instantiated anyhow.

        """
        if create and self._fileDialog is None:
            self._fileDialog = QFileDialog(self)
        return self._fileDialog

    def setPath(self, path):
        """Set the text in the lineEdit, without triggering any signals."""
        self._originalPath = path
        self.lineEdit.setText(path)

    def path(self):
        return self.lineEdit.text()

    def setFileMode(self, mode):
        """Sets the mode for the dialog, is a QFileDialog.FileMode value."""
        if mode == QFileDialog.FileMode.Directory:
            self.button.setIcon(icons.get('folder-open'))
        else:
            self.button.setIcon(icons.get('document-open'))
        self._fileMode = mode

    def fileMode(self):
        return self._fileMode

    def mustExist(self):
        """If True only existing files or directories are accepted as paths."""
        return self._mustExist

    def setMustExist(self, value):
        self._mustExist = value

    def setDialogTitle(self, title):
        self._dialogTitle = title
        if self._fileDialog:
            self._fileDialog.setWindowTitle(title)

    def dialogTitle(self):
        return self._dialogTitle

    def browse(self):
        """Opens the dialog."""
        # Suppress the editingFinished signal from the LineEdit
        self._browse_clicked = True
        dlg = self.fileDialog(True)
        dlg.setFileMode(self._fileMode)
        if self._dialogTitle:
            title = self._dialogTitle
        elif self.fileMode() == QFileDialog.FileMode.Directory:
            title = _("Select a directory")
        else:
            title = _("Select a file")
        dlg.setWindowTitle(app.caption(title))
        dlg.selectFile(self.path())
        result = dlg.exec()
        self._browse_clicked = False
        if result:
            new = dlg.selectedFiles()[0]
            self.lineEdit.setText(new)
            self._editingFinished()



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2013 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Maintains an overview of the structure of a Document.
"""


import re

from PyQt6.QtCore import QSettings

import app
import plugin

import lydocument
import ly.document

# default outline patterns that are ignored in comments
default_outline_patterns = [
r"(?P<title>\\(score|book|bookpart))\b",
r"^\\(paper|layout|header)\b",
r"\\(new|context)\s+[A-Z]\w+",
r"^[a-zA-Z]+\s*=",
r"^<<",
r"^\{",
r"^\\relative([ \t]+\w+[',]*)?",
]

# default outline patterns that are matched also in comments
default_outline_patterns_comments = [
r"(?P<title>BEGIN[^\n]*)[ \t]*$",
r"\b(?P<alert>(FIXME|HACK|XXX+)\b\W*\w+)",
]


# cache the outline regexp
_outline_re = None
_outline_re_comments = None


def outline_re(comments):
    """Return the expression to look for document outline items.
    If comments is True it is used to search in the whole document,
    if it is False comments are excluded."""
    v = '_outline_re'+('_comments' if comments else '')
    if globals()[v] is None:
        globals()[v] = create_outline_re(comments)
    return globals()[v]

def _reset_outline_re():
    global _outline_re
    global _outline_re_comments
    _outline_re = None
    _outline_re_comments = None


app.settingsChanged.connect(_reset_outline_re, -999)


def create_outline_re(comments):
    """Create and return the expression to look for document outline items.
    If comments is True it is used to search in the whole document,
    if it is False comments are excluded."""
    try:
        if comments:
            rx = QSettings().value("documentstructure/outline_patterns_comments",
                                   default_outline_patterns_comments, str)
        else:
            rx = QSettings().value("documentstructure/outline_patterns",
                                default_outline_patterns, str)
    except TypeError:
        rx = []
    # suffix duplicate named groups with a number
    groups = {}
    new_rx = []
    for e in rx:
        try:
            c = re.compile(e)
        except re.error:
            continue
        if c.groupindex:
            for name in c.groupindex:
                if name in groups:
                    groups[name] += 1
                    new_name = name + format(groups[name])
                    e = e.replace(f"(?P<{name}>", f"(?P<{new_name}>")
                else:
                    groups[name] = 0
        new_rx.append(e)
    rx = '|'.join(new_rx)
    return re.compile(rx, re.MULTILINE | re.UNICODE)


class DocumentStructure(plugin.DocumentPlugin):
    def __init__(self, document):
        self._outline = None

    def invalidate(self):
        """Called when the document changes or the settings are changed."""
        self._outline = None
        app.settingsChanged.disconnect(self.invalidate)
        self.document().contentsChanged.disconnect(self.invalidate)

    def outline(self):
        """Return the document outline as a series of match objects."""
        if self._outline is None:
            # match patterns excluding comments
            active_code = self.remove_comments()
            outline_list = list(outline_re(False).finditer(active_code))
            # match patterns including comments
            outline_list_comments = list(outline_re(True).finditer(self.document().toPlainText()))
            # merge lists and sort by start position
            self._outline = outline_list + outline_list_comments
            self._outline.sort(key=lambda match: match.start())
            self.document().contentsChanged.connect(self.invalidate)
            app.settingsChanged.connect(self.invalidate, -999)
        return self._outline

    def remove_comments(self):
        """Remove Lilypond comments from text"""
        def whiteout_section(cursor, start, end):
            spaces = ''.join(' ' for x in range(start, end))
            with cursor.document as doc:
                doc[start:end] = spaces

        doc = ly.document.Document(self.document().toPlainText())
        cursor = lydocument.Cursor(doc)
        source = ly.document.Source(cursor, True, tokens_with_position=True)
        start = 0
        for token in source:
            if isinstance(token, ly.lex.BlockCommentStart):
                start = token.pos
            elif isinstance(token, ly.lex.BlockCommentEnd):
                if start:
                    whiteout_section(cursor, start, token.end)
                    start = 0
            elif isinstance(token, ly.lex.Comment):
                whiteout_section(cursor, token.pos, token.end)
        return cursor.document.plaintext()




# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
The PDF preview panel widget.
"""



import collections
import itertools
import os

from PyQt6.QtCore import pyqtSignal, QMargins, QPoint, QRect, QSettings, Qt, QUrl
from PyQt6.QtGui import QCursor, QTextCharFormat
from PyQt6.QtWidgets import QToolTip, QVBoxLayout, QWidget

import qpageview
import qpageview.layout
import qpageview.highlight
import qpageview.rubberband
import pagedview

import app
import icons
import helpers
import textedit
import textformats
import lydocument
import viewhighlighter
import ly.lex.lilypond

from . import pointandclick


class MusicView(QWidget):
    """Widget containing the qpageview.View."""
    def __init__(self, dockwidget):
        """Creates the Music View for the dockwidget."""
        super().__init__(dockwidget)

        self._links = None
        self._clicking_link = False

        self._highlightFormat = QTextCharFormat()
        self._highlightMusicFormat = qpageview.highlight.Highlighter()
        self._highlightRange = None

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        self.view = pagedview.PagedView(self)
        self.view.setRubberband(qpageview.rubberband.Rubberband())
        layout.addWidget(self.view)
        app.settingsChanged.connect(self.readSettings)
        self.readSettings()
        self.view.setLinkHighlighter(qpageview.highlight.Highlighter())
        self.view.linkClicked.connect(self.slotLinkClicked)
        self.view.linkHovered.connect(self.slotLinkHovered)
        self.view.linkLeft.connect(self.slotLinkLeft)
        #self.view.setShowUrlTips(False)
        self.view.linkHelpRequested.connect(self.slotLinkHelpRequested)

        self.view.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.view.customContextMenuRequested.connect(self.showContextMenu)

        # react if cursor of current text document moves
        dockwidget.mainwindow().currentViewChanged.connect(self.slotCurrentViewChanged)
        view = dockwidget.mainwindow().currentView()
        if view:
            self.slotCurrentViewChanged(view)

    def sizeHint(self):
        """Returns the initial size the PDF (Music) View prefers."""
        return self.parent().mainwindow().size() / 2

    def openDocument(self, doc):
        """Open a qpageview.Document instance."""
        self._links = None
        self._highlightRange = None
        self._links = pointandclick.links(doc)
        self.view.setDocument(doc)

    def clear(self):
        """Empties the view."""
        self._links = None
        self._highlightRange = None
        self.view.clear()

    def readSettings(self):
        """Reads the settings from the user's preferences."""
        # background and highlight colors of music view
        colors = textformats.formatData('editor').baseColors
        self._highlightMusicFormat.color = colors['musichighlight']
        color = colors['selectionbackground']
        color.setAlpha(128)
        self._highlightFormat.setBackground(color)
        if QSettings().value("musicview/document_properties", True, bool):
            self.view.documentPropertyStore.mask = None
        else:
            self.view.documentPropertyStore.mask = ('position')

    def slotLinkClicked(self, ev, page, link):
        """Called when the use clicks a link.

        If the links is a textedit link, opens the document and puts the cursor there.
        Otherwise, call the helpers module to open the destination.

        """
        if ev.button() == Qt.MouseButton.RightButton:
            return
        cursor = self._links.cursor(link, True)
        if cursor:
            if ev.modifiers() & Qt.KeyboardModifier.ShiftModifier:
                import editinplace
                editinplace.edit(self, cursor, ev.globalPosition().toPoint())
            else:
                import browseriface
                mainwindow = self.parent().mainwindow()
                self._clicking_link = True
                browseriface.get(mainwindow).setTextCursor(cursor, findOpenView=True)
                self._clicking_link = False
                import widgets.blink
                widgets.blink.Blinker.blink_cursor(mainwindow.currentView())
                mainwindow.activateWindow()
                mainwindow.currentView().setFocus()
        elif link.url and not link.url.startswith('textedit:'):
            helpers.openUrl(QUrl(link.url))
        elif link.targetPage != -1 and not link.isExternal:
            self.view.setCurrentPageNumber(link.targetPage)

    def slotLinkHovered(self, page, link):
        """Called when the mouse hovers a link.

        If the links points to the current editor document, the token(s) it points
        at are highlighted using a transparent selection color.

        The highlight shows for a few seconds but disappears when the mouse moves
        off the link or when the link is clicked.

        """
        cursor = self._links.cursor(link)
        if not cursor or cursor.document() != self.parent().mainwindow().currentDocument():
            return

        # highlight token(s) at this cursor
        cursors = pointandclick.positions(cursor)
        if cursors:
            view = self.parent().mainwindow().currentView()
            viewhighlighter.highlighter(view).highlight(self._highlightFormat, cursors, 2, 5000)

    def slotLinkLeft(self):
        """Called when the mouse moves off a previously highlighted link."""
        view = self.parent().mainwindow().currentView()
        viewhighlighter.highlighter(view).clear(self._highlightFormat)

    def slotLinkHelpRequested(self, ev, page, link):
        """Called when a ToolTip wants to appear above the hovered link."""
        pos = self.view.viewport().mapToGlobal(ev.pos())
        if link.url:
            cursor = self._links.cursor(link)
            if cursor:
                import documenttooltip
                text = documenttooltip.text(cursor)
            else:
                l = textedit.link(link.url)
                if l:
                    text = f"{os.path.basename(l.filename)} ({l.line}:{l.column})"
                else:
                    text = link.url
        elif link.targetPage != -1:
            text = _("Page {num}").format(num=link.targetPage)
            if link.isExternal:
                text = link.fileName + "\n" + text
        else:
            return
        QToolTip.showText(pos, text, self.view, page.linkRect(link))

    def slotCurrentViewChanged(self, view, old=None):
        self.view.clearHighlight(self._highlightMusicFormat)
        if old:
            old.cursorPositionChanged.disconnect(self.slotCursorPositionChanged)
        view.cursorPositionChanged.connect(self.slotCursorPositionChanged)

    def slotCursorPositionChanged(self):
        """Called when the user moves the text cursor."""
        self.showCurrentLinks(
            not self._clicking_link
            and self.parent().actionCollection.music_sync_cursor.isChecked())

    def showCurrentLinks(self, scroll=False, msec=None):
        """Show objects at current text cursor.

        If scroll is True, also scrolls the view if needed. If msec is given,
        objects are highlighed that long. If not given, a default time is used.

        """
        if not self.isVisible() or not self._links:
            return # not visible of no PDF in the viewer

        view = self.parent().mainwindow().currentView()
        links = self._links.boundLinks(view.document())
        if not links:
            return # the PDF contains no references to the current text document

        s = links.indices(view.textCursor())
        if not s:
            if s is False:
                self.view.clearHighlight(self._highlightMusicFormat)
            return

        if not scroll and self._highlightRange == s and self.view.isHighlighting(self._highlightMusicFormat):
            return # don't redraw if same
        self._highlightRange = s

        # create the dict of areas
        layout = self.view.pageLayout()
        areas = collections.defaultdict(list)
        for dest in links.destinations()[s]:
            for pageNum, rect in dest:
                areas[layout[pageNum]].append(rect)

        if scroll:
            # compute the bounding rect
            boundingRect = QRect()
            for page, rects in areas.items():
                f = page.mapToPage(1, 1).rect
                pbound = QRect()
                for r in rects:
                    pbound |= f(r)
                boundingRect |= pbound.translated(page.pos())
            self.view.ensureVisible(boundingRect, QMargins(20, 20, 20, 20))

        if msec is None:
            msec = 5000 if s.stop - s.start > 1 else 2000 # show selections longer
        # RC: increased timer to give some time to the kinetic scrolling to complete.
        if self.view.isScrolling():
            msec += self.view.remainingScrollTime()
        self.view.highlight(areas, self._highlightMusicFormat, msec)

    def showContextMenu(self):
        """Called when the user right-clicks or presses the context menu key."""
        pos = self.view.mapToGlobal(QPoint(0, 0))
        link, cursor = None, None
        # mouse inside view?
        if self.view.mapFromGlobal(QCursor.pos()) in self.view.viewport().rect():
            pos = QCursor.pos()
            pos_in_viewport = self.view.viewport().mapFromGlobal(pos)
            page, link = self.view.linkAt(pos_in_viewport)
            if link:
                cursor = self._links.cursor(link, True)
        from . import contextmenu
        contextmenu.show(pos, self.parent(), link, cursor)



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
LilyPond preferences page
"""


from dataclasses import dataclass
import functools
import json
import os
import platform
import shutil
import sys
import tempfile

from PyQt6.QtCore import pyqtSignal, QSettings, Qt, QTimer, QUrl
from PyQt6.QtNetwork import QNetworkAccessManager, QNetworkRequest
from PyQt6.QtGui import QAction
from PyQt6.QtWidgets import (
    QAbstractItemView, QCheckBox, QDialog, QDialogButtonBox,
    QFileDialog, QGridLayout, QHBoxLayout, QLabel, QLineEdit, QListWidgetItem,
    QMenu, QMessageBox, QPushButton, QRadioButton, QTabWidget, QVBoxLayout,
    QWidget)

import app
import userguide
import qutil
import icons
import preferences
import lilypondinfo
import linux
import qsettings
import widgets.listedit
import widgets.urlrequester


def settings():
    s = QSettings()
    s.beginGroup("lilypond_settings")
    return s

@functools.lru_cache(maxsize=None) # can become @functools.cache in Python 3.9+
def _network_manager():
    return QNetworkAccessManager()


class LilyPondPrefs(preferences.ScrolledGroupsPage):
    def __init__(self, dialog):
        super().__init__(dialog)

        layout = QVBoxLayout()
        self.scrolledWidget.setLayout(layout)

        layout.addWidget(Versions(self))
        layout.addWidget(Target(self))
        layout.addWidget(Running(self))


class Versions(preferences.Group):
    def __init__(self, page):
        super().__init__(page)

        layout = QVBoxLayout()
        self.setLayout(layout)

        self.instances = InfoList(self)
        self.instances.changed.connect(self.changed)
        self.instances.defaultButton.clicked.connect(self.defaultButtonClicked)
        layout.addWidget(self.instances)
        self.autoVersion = QCheckBox(clicked=self.changed)
        layout.addWidget(self.autoVersion)
        app.translateUI(self)
        userguide.openWhatsThis(self)

    def defaultButtonClicked(self):
        item = self.instances.listBox.currentItem()
        # This holds because otherwise the "Set as Default" button would
        # be greyed out.
        assert isinstance(item.state, InstalledState)
        # Register the current default instance (identified by its command)
        self._defaultCommand = item.state.info.command
        # Remove the [default] marker from the previous item that had it.
        for item in self.instances.items():
            item.display()
        self.changed.emit()

    def translateUI(self):
        self.setTitle(_("LilyPond versions to use"))
        self.autoVersion.setText(_("Automatically choose LilyPond version from document"))
        self.autoVersion.setToolTip(_(
            "If checked, the document's version determines the LilyPond version to use.\n"
            "See \"What's This\" for more information."))
        self.autoVersion.setWhatsThis(userguide.html("prefs_lilypond_autoversion") +
            _("See also {link}.").format(link=userguide.link("prefs_lilypond")))

    def loadSettings(self):
        s = settings()
        default = lilypondinfo.default()
        self._defaultCommand = s.value("default", default.command, str)
        self.autoVersion.setChecked(s.value("autoversion", False, bool))
        infos = sorted(lilypondinfo.infos(), key=lambda i: i.version())
        if not infos:
            infos = [default]
        items = [InfoItem(InstalledState(info)) for info in infos]
        self.instances.setItems(items)
        # Make the default version selected initially
        for item in items:
            if item.state.info.command == self._defaultCommand:
                self.instances.setCurrentItem(item)
                break

    def saveSettings(self):
        infos = []
        pending_items = []
        for item in self.instances.items():
            if isinstance(item.state, DownloadingState):
                pending_items.append(item)
            elif isinstance(item.state, InstalledState):
                infos.append(item.state.info)
            # nothing to do for DownloadErrorState, just discard these
        if pending_items:
            answer = QMessageBox.warning(
                self, app.caption("Downloads pending"),
                _("LilyPond downloads are in progress. Stop them?"),
                QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel)
            if answer == QMessageBox.StandardButton.Cancel:
                raise preferences.CancelClosingPreferences
            else:
                # Stop unfinished downloads
                for item in pending_items:
                    item.abortDownload()
        if infos:
            for info in infos:
                if info.command == self._defaultCommand:
                    break
            else:
                self._defaultCommand = infos[0].command
        else:
            infos = [lilypondinfo.default()]
            self._defaultCommand = infos[0].command
        s = settings()
        s.setValue("default", self._defaultCommand)
        s.setValue("autoversion", self.autoVersion.isChecked())
        lilypondinfo.setinfos(infos)
        lilypondinfo.saveinfos()


class InfoList(widgets.listedit.ListEdit):
    def __init__(self, group):
        # Compared to a basic ListEdit, this widget also has a "Set as Default"
        # button. This must be initialized before the base class __init__, because
        # it calls updateSelection().
        self.defaultButton = QPushButton()

        super().__init__(group)

        # The "Add" button is transformed into a menu.
        self.addMenu = AddMenu(self)
        self.addButton.setMenu(self.addMenu)

        self.layout().addWidget(self.defaultButton, 3, 1)
        self.layout().addWidget(self.listBox, 0, 0, 5, 1)

    def translateUI(self):
        super().translateUI()
        self.defaultButton.setText(_("Set as &Default"))

    def updateSelection(self):
        """Overridden to grey out the "Edit" and "Set as Default" buttons in more cases."""
        super().updateSelection()
        # "Edit" and "Set as Default" are greyed out not only when there are
        # no items at all, but also when the current item is for a version that
        # is being downloaded. In theory, "Edit" could be made to work (with
        # some refactorings) on such items, but the use case for auto-managed
        # versions is not clear. Likewise, we could make "Set as Default" work,
        # but it is problematic to know what to do if a version is set as
        # default while being downloaded, and then the download fails (which
        # becomes the default afterwards?). Practically, not being able to
        # edit an installation that is being downloaded is probably intuitively
        # normal for most people.
        item = self.listBox.currentItem()
        is_installed_item = bool(item) and isinstance(item.state, InstalledState)
        self.defaultButton.setEnabled(is_installed_item)
        self.editButton.setEnabled(is_installed_item)

    def createItem(self):
        if linux.inside_flatpak():
            QMessageBox.warning(self, app.caption(_("Warning")),
                _("You are running a Flatpak installation of Frescobaldi. Use this menu "
                  "only to add a LilyPond installation in /app (flatpak bundle) that "
                  "was accidentally removed or not detected by Frescobaldi.\n\n"
                  "Because Flatpak isolates apps from the rest of the system, custom LilyPond "
                  "binaries are not supported. Please either let Frescobaldi auto-install "
                  "a given LilyPond version, using the “Add…” menu, or if you wish to use "
                  "a custom build, install Frescobaldi from a distribution package or from source."))
        return InfoItem(InstalledState(lilypondinfo.LilyPondInfo("lilypond")))

    def openEditor(self, item):
        if item is None: # Flatpak case above
            return False

        # Either we're adding a custom version and the item is created in createItem
        # above, or we're editing an existing version because the "Edit" button was clicked,
        # and it's disabled for versions that are being downloaded or failed to download.
        # Either way, we can only edit fully installed versions.
        assert isinstance(item.state, InstalledState)

        dlg = InfoDialog(self, item.state.info)
        # Since we use the command as the way to remember which version is the
        # default, and the user may edit the command here, we need to update
        # it. The check "item in self.items()" is true if this is an existing
        # item that we are updating, with the "Edit" button, but false if it's
        # a fresh item as created by the "Add" button in createItem() above.
        was_default = (item in self.items()
                       and item.state.info.command == self.parentWidget()._defaultCommand)
        if dlg.exec():
            item.state.info = dlg.newInfo()
            if was_default:
                self.parentWidget()._defaultCommand = item.state.info.command
            return True
        return False

    def itemChanged(self, item):
        item.display()
        self.setCurrentItem(item)


    def itemRemoved(self, item):
        if isinstance(item.state, DownloadingState):
            # Cancel a pending download
            item.abortDownload()
        elif isinstance(item.state, InstalledState):
            # Wipe out if auto-managed
            item.state.info.forget()
        # If this version of LilyPond is no longer installed, the menu needs to know
        # about it, to reenable the action.
        self.addButton.menu().updateActions()



class ItemState:
    """Base class for classes holding the state of an InfoItem."""

@dataclass
class DownloadingState(ItemState):
    """Data for an InfoItem whose LilyPond is being downloaded from the network."""
    version_string: str
    download_progress: int  = 0 # percentage

@dataclass
class DownloadErrorState(ItemState):
    """Data for an InfoItem whose LilyPond could not be downloaded."""
    version_string: str

@dataclass
class InstalledState(ItemState):
    """Data for an InfoItem whose LilyPond is fully installed on the system."""
    info: lilypondinfo.LilyPondInfo


_LILYPOND_PACKAGE_URL = (
    "https://gitlab.com/api/v4/projects/lilypond%2Flilypond/packages/"
    + "generic/lilypond/{version_string}/{archive}"
)

class InfoItem(QListWidgetItem):
    """An item in the list of LilyPond versions.

    The `state` attribute holds an instance of `ItemState`.
    """

    # TODO: Currently, download progress is displayed with a percentage in text,
    # but it would be very nice to have it shown as a translucent progress bar
    # overlaid behind the item.

    def __init__(self, state):
        super().__init__()
        self.state = state

    @classmethod
    def fromNetwork(cls, infolist, version):
        """Make an InfoItem that will auto-download and install a given LilyPond version."""
        major, minor, micro = version
        version_string = f"{major}.{minor}.{micro}"
        self = cls(DownloadingState(version_string))

        self.parent = infolist

        if platform.system() == 'Linux':
            archive = f"lilypond-{major}.{minor}.{micro}-linux-x86_64.tar.gz"
            self.archive_format = "gztar"
        elif platform.system() == 'Darwin':
            archive = f"lilypond-{major}.{minor}.{micro}-darwin-x86_64.tar.gz"
            self.archive_format = "gztar"
        elif platform.system() == 'Windows':
            archive = f"lilypond-{major}.{minor}.{micro}-mingw-x86_64.zip"
            self.archive_format = "zip"
        else:
            raise AssertionError

        # Download an archive of LilyPond binaries from GitLab
        archive_url = QUrl(_LILYPOND_PACKAGE_URL.format(
                             version_string=version_string, archive=archive))

        self.reply = _network_manager().get(QNetworkRequest(archive_url))
        self.reply.downloadProgress.connect(self.updateProgress)
        self.reply.finished.connect(self.downloadFinished)
        self.reply.errorOccurred.connect(self.downloadError)

        if platform.system() == "Windows":
            exec_name = "lilypond.exe"
        else:
            exec_name = "lilypond"
        self.executable_path = os.path.join(lilypondinfo.LILYPOND_AUTOINSTALL_DIR,
                                            f"lilypond-{version_string}", "bin",
                                            exec_name)
        return self

    def updateProgress(self, bytes_received, bytes_total):
        if isinstance(self.state, DownloadErrorState):
            # This happens when the download is cancelled with abortDownload();
            # Qt triggers the downloadProgress signal one last time.
            return
        progress = round(bytes_received/bytes_total * 100)
        assert isinstance(self.state, DownloadingState)
        self.state.download_progress = progress
        self.display()

    def downloadFinished(self):
        if isinstance(self.state, DownloadErrorState):
            return # the finished signal is called even in the error case
        assert isinstance(self.state, DownloadingState)
        # Unpack the downloaded archive
        with tempfile.NamedTemporaryFile(delete=False) as tfile:
            tfile.write(self.reply.readAll())
            tfile.close()   # needed to not block unpack_archive() on Windows
            shutil.unpack_archive(tfile.name, lilypondinfo.LILYPOND_AUTOINSTALL_DIR,
                                  format=self.archive_format)
            # On Python >= 3.12 we can create tfile with delete=True and
            # delete_on_close=False to avoid having to do this manually,
            # but this is not backwards compatible with older versions
            os.unlink(tfile.name)
        # Switch to "installed" state
        self.state = InstalledState(lilypondinfo.LilyPondInfo(self.executable_path))
        self.display()
        # If this item is still selected, the parent will need to reenable
        # the "Edit" and "Set as Default" buttons.
        self.parent.updateSelection()

    def downloadError(self, code):
        assert isinstance(self.state, DownloadingState)
        self.state = DownloadErrorState(self.state.version_string)
        self.display()

    def abortDownload(self):
        self.reply.abort() # also calls downloadError()

    def display(self):
        if isinstance(self.state, DownloadErrorState):
            icon = "dialog-error"
            text = _("Could not download LilyPond {version}").format(
                       version=self.state.version_string)
        elif isinstance(self.state, DownloadingState):
            icon = "view-refresh"
            text = _("Downloading LilyPond {version}... {progress}%").format(
                      version=self.state.version_string,
                      progress=self.state.download_progress)
        elif isinstance(self.state, InstalledState):
            if self.state.info.version():
                icon = "lilypond-run"
            else:
                icon = "dialog-error"
            text = self.state.info.prettyName()
            default = self.listWidget().parentWidget().parentWidget()._defaultCommand
            if self.state.info.command == default:
                text += " [{}]".format(_("default"))

        self.setIcon(icons.get(icon))
        self.setText(text)


_LILYPOND_RELEASES_API_URL = "https://gitlab.com/api/v4/projects/18695663/releases"

class AddMenu(QMenu):
    """Menu added to the "Add" button of the version list.

    It offers several methods of adding: auto-download, or use a local install.
    There is a set of actions for auto-download and one "Custom" action.
    Initially, the auto-download part displays a greyed-out action while
    downloading the list of available LilyPond versions. When the download
    completes, the normal buttons are added. If the download fails, the
    greyed-out action is replaced with another one to signal the failure.
    """
    def __init__(self, parent):
        super().__init__(parent)
        # Fake menu item as placeholder while the download is in progress.
        self.addAction(_("Downloading LilyPond version list...")).setEnabled(False)
        # Action to add custom LilyPond version anywhere on the file system.
        self.customAction = QAction(_("Button to register a custom LilyPond installation", "Custom..."))
        self.addAction(self.customAction)
        # In most ListEdit's, this is the slot connected to the "Add"
        # button, here's it's connected to the "Custom" button.
        self.customAction.triggered.connect(parent.addClicked)

        # Get the list of available LilyPond versions using the GitLab API.
        url = QUrl(_LILYPOND_RELEASES_API_URL)
        self.reply = _network_manager().get(QNetworkRequest(url))
        self.reply.finished.connect(self.downloadFinished)
        self.reply.errorOccurred.connect(self.downloadError)
        self.hasError = False
        self.downloadSuccessful = False

    def downloadFinished(self):
        if self.hasError:
            return # self.reply.finished is triggered even in case of error

        self.downloadSuccessful = True

        # Decode the JSON data returned by the API
        data = json.loads(bytes(self.reply.readAll()))
        versions = []
        for item in data:
            tag = item["tag_name"]
            # This release dates back to before LilyPond binaries became simple
            # for us to install. (Its tag also uses a different naming convention.)
            if tag == "release/2.22.2-1":
                continue
            # tag has the form v2.x.y
            version = tuple(int(n) for n in tag.lstrip("v").split("."))
            versions.append(version)
        versions.sort(reverse=True) # most recent version first
        self.clear() # rebuild the menu
        self.actions = []
        # Find the latest stable version
        for (major, minor, micro) in versions:
            if minor % 2 == 0:
                latest_stable = (major, minor, micro)
                stable_text = (_("latest stable LilyPond release",
                                 "{major}.{minor}.{micro} (latest &stable)")
                               .format(major=major, minor=minor, micro=micro))
                # Make a dedicated menu item for this version, to make it
                # reachable very easily.
                self.makeVersionAction(latest_stable, stable_text, self)
                versions.remove(latest_stable)
                break
        else: # no `break` reached
            raise AssertionError # at least 2.24 should be in the list
        # Likewise, find the latest unstable version and add a menu item
        for (major, minor, micro) in versions:
            if minor % 2 == 1:
                latest_unstable = (major, minor, micro)
                unstable_text = (_("latest unstable LilyPond release",
                                   "{major}.{minor}.{micro} (latest &unstable)")
                                 .format(major=major, minor=minor, micro=micro))
                self.makeVersionAction(latest_unstable, unstable_text, self)
                versions.remove(latest_unstable)
                break
        else:
            raise AssertionError
        # Add a submenu for all other versions.
        other_text = _("LilyPond versions other than latest stable/unstable",
                       "Other...")
        self.otherMenu = self.addMenu(other_text)
        for version in versions:
            major, minor, micro = version
            text = f"{major}.{minor}.{micro}"
            self.makeVersionAction(version, text, self.otherMenu)

        # Add the "Custom" action that we already had.
        self.addAction(self.customAction)

        # Update which actions are enabled
        self.updateActions()

    def makeVersionAction(self, version, text, parent):
        action = QAction(text, parent)
        parent.addAction(action)
        # Connect the action
        action.triggered.connect(self.autoDownload(version))
        self.actions.append((version, action))

    def downloadError(self, code):
        self.hasError = True
        self.clear() # rebuild the menu
        self.addAction(_("Could not download LilyPond version list")).setEnabled(False)
        self.addAction(self.customAction)

    def autoDownload(self, version):
        """Return a function that downloads a certain version, for connecting to menu actions."""
        def download_inner():
            item = InfoItem.fromNetwork(self.parent(), version)
            self.parent().addItem(item)
            self.updateActions()
        return download_inner

    def updateActions(self):
        """
        Disable a menu action if there is already an auto-managed LilyPond installation for the same version.

        Prevents the user from doing something silly and spares us error
        handling downstream to cope with duplicate LilyPondInfo's for the same
        installation (e.g., when removing them).
        """
        if not self.downloadSuccessful:
            # We aren't done downloading the list of versions yet (or there was
            # an error), so we don't have the version actions and we needn't care
            # about this.
            return

        disabled_versions = set()
        for item in self.parent().items():
            if isinstance(item.state, InstalledState) and item.state.info.isAutoManaged:
                disabled_versions.add(item.state.info.versionString())
            elif isinstance(item.state, DownloadingState):
                disabled_versions.add(item.state.version_string)

        for (major, minor, micro), action in self.actions:
            enable = (f"{major}.{minor}.{micro}" not in disabled_versions)
            action.setEnabled(enable)


class InfoDialog(QDialog):
    def __init__(self, parent, info):
        super().__init__(parent)
        self.setWindowModality(Qt.WindowModality.WindowModal)

        layout = QVBoxLayout()
        layout.setSpacing(10)
        self.setLayout(layout)

        self.tab = QTabWidget()
        tab_general = QWidget()
        tab_toolcommands = QWidget()
        self.tab.addTab(tab_general, "")
        self.tab.addTab(tab_toolcommands, "")

        # general tab
        vbox = QVBoxLayout()
        vbox.setSpacing(4)
        tab_general.setLayout(vbox)

        hbox = QHBoxLayout()
        self.lilyname = QLineEdit()
        self.lilyname.setText(info.name)
        self.lilynameLabel = l = QLabel()
        l.setBuddy(self.lilyname)
        hbox.addWidget(l)
        hbox.addWidget(self.lilyname)
        vbox.addLayout(hbox)

        # For auto-managed installation, don't show the installation
        # path to the user or let them edit it.
        self.haveCommandEdit = not info.isAutoManaged
        if self.haveCommandEdit:
            self.lilypond = widgets.urlrequester.UrlRequester()
            self.lilypond.setFileMode(QFileDialog.FileMode.ExistingFile)
            self.lilypond.setPath(info.command)
            self.lilypond.lineEdit.setFocus()
            self.lilypondLabel = l = QLabel()
            l.setBuddy(self.lilypond)
            vbox.addWidget(l)
            vbox.addWidget(self.lilypond)
        else:
            self.lilypondCommand = info.command

        self.autoVersionIncluded = QCheckBox()
        self.autoVersionIncluded.setChecked(info.autoVersionIncluded)
        vbox.addWidget(self.autoVersionIncluded)
        vbox.addStretch(1)

        # toolcommands tab
        grid = QGridLayout()
        grid.setSpacing(4)
        tab_toolcommands.setLayout(grid)

        self.ly_tool_widgets = {}
        row = 0
        for name, gui in self.toolnames():
            w = QLineEdit()
            l = QLabel()
            l.setBuddy(w)
            grid.addWidget(l, row, 0)
            grid.addWidget(w, row, 1)
            row += 1
            self.ly_tool_widgets[name] = (l, w)
            self.ly_tool_widgets[name][1].setText(info.ly_tool(name))

        layout.addWidget(self.tab)
        layout.addWidget(widgets.Separator())
        b = self.buttons = QDialogButtonBox(self)
        layout.addWidget(b)

        b.setStandardButtons(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        b.accepted.connect(self.accept)
        b.rejected.connect(self.reject)
        userguide.addButton(b, "prefs_lilypond")
        app.translateUI(self)
        qutil.saveDialogSize(self, "/preferences/lilypond/lilypondinfo/dialog/size")

    def toolnames(self):
        """Yield tuples (name, GUI name) for the sub tools we allow to be configured."""
        yield 'convert-ly', _("Convert-ly:")
        yield 'lilypond-book', _("LilyPond-book:")
        yield 'midi2ly', _("Midi2ly:")
        yield 'musicxml2ly', _("MusicXML2ly:")
        yield 'abc2ly', _("ABC2ly:")

    def translateUI(self):
        self.setWindowTitle(app.caption(_("LilyPond")))
        self.lilynameLabel.setText(_("Label:"))
        self.lilynameLabel.setToolTip(_("How this version of LilyPond will be displayed."))
        if self.haveCommandEdit:
            self.lilypondLabel.setText(_("LilyPond Command:"))
            self.lilypond.lineEdit.setToolTip(_("Name or full path of the LilyPond program."))
        self.autoVersionIncluded.setText(_("Include in automatic version selection"))
        self.tab.setTabText(0, _("General"))
        self.tab.setTabText(1, _("Tool Commands"))
        for name, gui in self.toolnames():
            self.ly_tool_widgets[name][0].setText(gui)

    def newInfo(self):
        """Returns a new LilyPondInfo instance for our settings."""
        if not self.haveCommandEdit:
            info = lilypondinfo.LilyPondInfo(self.lilypondCommand)
        elif platform.system() == "Darwin" and self.lilypond.path().endswith('.app'):
            info = lilypondinfo.LilyPondInfo(
                self.lilypond.path() + '/Contents/Resources/bin/lilypond')
        else:
            info = lilypondinfo.LilyPondInfo(self.lilypond.path())
        if self.lilyname.text() and not self.lilyname.text().isspace():
            info.name = self.lilyname.text()
        info.autoVersionIncluded = self.autoVersionIncluded.isChecked()
        for name, gui in self.toolnames():
            info.set_ly_tool(name, self.ly_tool_widgets[name][1].text())
        return info


class Running(preferences.Group):
    def __init__(self, page):
        super().__init__(page)

        layout = QVBoxLayout()
        self.setLayout(layout)

        self.saveDocument = QCheckBox(clicked=self.changed)
        self.deleteFiles = QCheckBox(clicked=self.changed)
        self.embedSourceCode = QCheckBox(clicked=self.changed)
        self.noTranslation = QCheckBox(clicked=self.changed)
        self.includeLabel = QLabel()
        self.include = widgets.listedit.FilePathEdit()
        self.include.listBox.setDragDropMode(
            QAbstractItemView.DragDropMode.InternalMove)
        self.include.changed.connect(self.changed)
        layout.addWidget(self.saveDocument)
        layout.addWidget(self.deleteFiles)
        layout.addWidget(self.embedSourceCode)
        layout.addWidget(self.noTranslation)
        layout.addWidget(self.includeLabel)
        layout.addWidget(self.include)
        app.translateUI(self)

    def translateUI(self):
        self.setTitle(_("Running LilyPond"))
        self.saveDocument.setText(_("Save document if possible"))
        self.saveDocument.setToolTip(_(
            "If checked, the document is saved when it is local and modified.\n"
            "Otherwise a temporary file is used to run LilyPond."))
        self.deleteFiles.setText(_("Delete intermediate output files"))
        self.deleteFiles.setToolTip(_(
            "If checked, LilyPond will delete intermediate PostScript files."))
        self.embedSourceCode.setText(_("Embed Source Code files in publish mode"))
        self.embedSourceCode.setToolTip(_(
            "If checked, the LilyPond source files will be embedded in the PDF\n"
            "when LilyPond is started in publish mode."))
        self.noTranslation.setText(_("Run LilyPond with English messages"))
        self.noTranslation.setToolTip(_(
            "If checked, LilyPond's output messages will be in English.\n"
            "This can be useful for bug reports."))
        self.includeLabel.setText(_("LilyPond include path:"))

    def loadSettings(self):
        s = settings()
        self.saveDocument.setChecked(s.value("save_on_run", False, bool))
        self.deleteFiles.setChecked(s.value("delete_intermediate_files", True, bool))
        self.embedSourceCode.setChecked(s.value("embed_source_code", False, bool))
        self.noTranslation.setChecked(s.value("no_translation", False, bool))
        include_path = qsettings.get_string_list(s, "include_path")
        self.include.setValue(include_path)

    def saveSettings(self):
        s = settings()
        s.setValue("save_on_run", self.saveDocument.isChecked())
        s.setValue("delete_intermediate_files", self.deleteFiles.isChecked())
        s.setValue("embed_source_code", self.embedSourceCode.isChecked())
        s.setValue("no_translation", self.noTranslation.isChecked())
        s.setValue("include_path", self.include.value())


class Target(preferences.Group):
    def __init__(self, page):
        super().__init__(page)

        layout = QGridLayout()
        self.setLayout(layout)

        self.targetPDF = QRadioButton(toggled=page.changed)
        self.targetSVG = QRadioButton(toggled=page.changed)
        self.openDefaultView = QCheckBox(clicked=page.changed)

        layout.addWidget(self.targetPDF, 0, 0)
        layout.addWidget(self.targetSVG, 0, 1)
        layout.addWidget(self.openDefaultView, 1, 0, 1, 5)
        app.translateUI(self)

    def translateUI(self):
        self.setTitle(_("Default output format"))
        self.targetPDF.setText(_("PDF"))
        self.targetPDF.setToolTip(_(
            "Create PDF (Portable Document Format) documents by default."))
        self.targetSVG.setText(_("SVG"))
        self.targetSVG.setToolTip(_(
            "Create SVG (Scalable Vector Graphics) documents by default."))
        self.openDefaultView.setText(_("Open default viewer after successful compile"))
        self.openDefaultView.setToolTip(_(
            "Shows the PDF or SVG music view when a compile job finishes "
            "successfully."))

    def loadSettings(self):
        s = settings()
        target = s.value("default_output_target", "pdf", str)
        if target == "svg":
            self.targetSVG.setChecked(True)
            self.targetPDF.setChecked(False)
        else:
            self.targetSVG.setChecked(False)
            self.targetPDF.setChecked(True)
        self.openDefaultView.setChecked(s.value("open_default_view", True, bool))

    def saveSettings(self):
        s = settings()
        if self.targetSVG.isChecked():
            target = "svg"
        else:
            target = "pdf"
        s.setValue("default_output_target", target)
        s.setValue("open_default_view", self.openDefaultView.isChecked())



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2015 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
A Dialog to run an external command and simply show the log.
"""

from PyQt6.QtCore import Qt, QSize

import log
import qutil
import signals
import widgets.dialog

class Dialog(widgets.dialog.Dialog):
    """Dialog to run arbitrary external job.Job commands and show the log."""

    job_done = signals.Signal()

    def __init__(self, parent, auto_accept=False):
        super().__init__(
            parent,
            buttons=('cancel', 'ok',)
        )
        self.setWindowModality(Qt.WindowModality.WindowModal)
        self.setIconSize(32)
        self.auto_accept = auto_accept
        self.log = log.Log(self)
        self.job = None
        self.setMainWidget(self.log)
        qutil.saveDialogSize(self, "job-dialog/dialog/size", QSize(480, 800))

    def abort_job(self):
        self.job.abort()
        self.reject()

    def run(self, j, msg=_("Run external command")):
        """Run the given job."""
        self.job = j
        self.job.done.connect(self.slot_job_done)
        self.log.connectJob(j)
        self.setWindowTitle(j.title())
        self.setMessage(msg)
        self.button('ok').setEnabled(False)
        self.button('cancel').clicked.connect(self.abort_job)
        j.start()
        self.exec()

    def slot_job_done(self):
        if self.job.success:
            if self.auto_accept:
                self.accept()
            self.button('ok').setEnabled(True)
        else:
            self.setMessage(_("Job failed! Please inspect log"))
            self.setIcon('critical')
        self.button('cancel').clicked.connect(self.reject)
        self.job_done.emit()
